#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from selenium.common.exceptions import StaleElementReferenceException, UnexpectedAlertPresentException, \
    InvalidSelectorException, WebDriverException, SessionNotCreatedException
from selenium.webdriver.common.by import By
from holado_ui.ui.gui.drivers.gui_internal_api import GUIInternalAPI
from holado_ui_selenium.ui.gui.selenium.drivers.selenium_driver import SeleniumDriver
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions.element_exception import NoSuchElementException,\
    TooManyElementsException
from holado_core.common.finders.tools.find_context import WithRootContainerFindContext,\
    ContainerFindContext, ListContainersFindContext
from holado_ui_selenium.ui.gui.selenium.handlers.selenium_by import SeleniumBy
import html
from selenium.webdriver.common.action_chains import ActionChains
from holado_core.common.exceptions.functional_exception import FunctionalException
from selenium.webdriver.support.select import Select
import os
from holado_core.common.tools.tools import Tools
import time
from selenium.webdriver.remote.webelement import WebElement
from Screenshot import Screenshot
from holado_core.common.tools.converters.converter import Converter
from holado_core.common.handlers.redo import Redo
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support import expected_conditions as EC
from holado_ui_selenium.ui.gui.selenium.criterias.selenium_criteria import SeleniumCriteria
from holado_python.standard_library.typing import Typing
from holado_python.common.tools.datetime import DateTime

logger = logging.getLogger(__name__)



class SeleniumInternalAPI(GUIInternalAPI):
    """ Implement all needed internal API methods with Selenium.
    """
    
    @staticmethod
    def selenium_version():
        """
        @return Selenium version
        """
        import selenium
        return selenium.__version__
    
    def __init__(self, driver:SeleniumDriver):
        super().__init__(driver)
        
        # Wait settings
        self.__wait_jquery = True
        self.__wait_angular = True
    
    @property
    def wait_angular(self):
        return self.__wait_angular
    
    @wait_angular.setter
    def wait_angular(self, wait_angular):
        """
        @param wait_angular If wait angular
        """
        self.__wait_angular = wait_angular
    
    @property
    def wait_jquery(self):
        return self.__wait_jquery
    
    @wait_jquery.setter
    def wait_jquery(self, wait_jquery):
        """
        @param wait_jquery If wait jquery
        """
        self.__wait_jquery = wait_jquery
        
    def find_element_by(self, by:SeleniumBy, criteria, find_context, find_parameters):
        """
        Find element with given criteria
        @param by By criteria
        @param criteria Criteria
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found element
        """
        if isinstance(find_context, ContainerFindContext) and find_context.container is not None:
            return self.find_element_by_container_by(by, criteria, find_context.container, find_context, find_parameters)
        elif isinstance(find_context, ListContainersFindContext) and find_context.nb_containers > 0:
            return self.find_element_by_candidates_by(by, criteria, find_context.containers, find_context, find_parameters)
        else:
            raise TechnicalException("Find context has no container")
    
    def find_elements_by(self, by:SeleniumBy, criteria, find_context, find_parameters):
        """
        Find elements in context container with given criteria
        @param by By criteria
        @param criteria Criteria
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found elements
        """
        if isinstance(find_context, ContainerFindContext) and find_context.container is not None:
            return self.find_elements_by_container_by(by, criteria, find_context.container, find_context, find_parameters)
        elif isinstance(find_context, ListContainersFindContext) and find_context.nb_containers > 0:
            return self.find_elements_by_candidates_by(by, criteria, find_context.containers, find_context, find_parameters)
        else:
            raise TechnicalException("Find context has no container")
    
    def find_element_by_container(self, criteria, container, find_context, find_parameters):
        if criteria is None:
            raise TechnicalException("Criteria must be defined")
        
        if isinstance(criteria, SeleniumCriteria):
            if criteria.class_name is not None:
                return self.find_element_by_container_by(SeleniumBy(By.CLASS_NAME, criteria.class_name), criteria, container, find_context, find_parameters)
            elif criteria.id is not None:
                return self.find_element_by_container_by(SeleniumBy(By.ID, criteria.id), criteria, container, find_context, find_parameters)
            elif criteria.name is not None:
                return self.find_element_by_container_by(SeleniumBy(By.NAME, criteria.name), criteria, container, find_context, find_parameters)
            elif criteria.getTagName() is not None:
                return self.find_element_by_container_by(SeleniumBy(By.TAG_NAME, criteria.tag_name), criteria, container, find_context, find_parameters)
            else:
                return self.find_element_in_tree(criteria, container, find_context, find_parameters)
        else:
            return self.find_element_in_tree(criteria, container, find_context, find_parameters)
        
    def find_element_by_container_by(self, by:SeleniumBy, criteria, container, find_context, find_parameters):
        """
        Find element in given container with given criteria
        @param by By criteria
        @param criteria Criteria
        @param container Container
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found element
        """
        res = None
        
        # Find elements
        found_elements = self.find_elements_by_container_by(by, criteria, container, find_context, find_parameters)
        
        # Analyse search result
        if len(found_elements) == 1:
            res = found_elements[0]
        elif len(found_elements) > 1:
            msg_list = ["More than one"]
            if find_parameters.nb_max_elements > 0:
                msg_list.append(f" (at least {len(found_elements)})")
            else:
                msg_list.append(f" ({len(found_elements)})")
            msg_list.append(f" element were found with: by={by}")
            if criteria is not None:
                msg_list.append(f" ; criteria={criteria}")
            msg_list.append(":\n")
            msg_list.append(self.represent_elements(found_elements, 4))
            raise TooManyElementsException("".join(msg_list))
        elif find_parameters.raise_no_such_element:
            if criteria is not None:
                raise NoSuchElementException(f"Unable to find element with: by={{{by}}} ; criteria={{{criteria}}} ; container={{{container.get_complete_description_and_details()}}}")
            else:
                raise NoSuchElementException(f"Unable to find element with: by={{{by}}} ; container={{{container.get_complete_description_and_details()}}}")
        
        return res
    
    def find_elements_by_container(self, criteria, container, find_context, find_parameters):
        if criteria is None:
            raise TechnicalException("Criteria must be defined")
        
        if isinstance(criteria, SeleniumCriteria):
            if criteria.class_name is not None:
                return self.find_elements_by_container_by(SeleniumBy(By.CLASS_NAME, criteria.class_name), criteria, container, find_context, find_parameters)
            elif criteria.id is not None:
                return self.find_elements_by_container_by(SeleniumBy(By.ID, criteria.id), criteria, container, find_context, find_parameters)
            elif criteria.name is not None:
                return self.find_elements_by_container_by(SeleniumBy(By.NAME, criteria.name), criteria, container, find_context, find_parameters)
            elif criteria.getTagName() is not None:
                return self.find_elements_by_container_by(SeleniumBy(By.TAG_NAME, criteria.tag_name), criteria, container, find_context, find_parameters)
            else:
                return self.find_elements_in_tree(criteria, container, find_context, find_parameters)
        else:
            return self.find_elements_in_tree(criteria, container, find_context, find_parameters)
        
    def find_elements_by_container_by(self, by:SeleniumBy, criteria, container, find_context, find_parameters):
        """
        Find elements in given container with given criteria
        @param by By criteria
        @param criteria Criteria
        @param container Container
        @param find_context Find context
        @param find_parameters Find parameters
        @return Found elements
        """
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"find_elements_by_container_by: by={{{by}}} ; criteria={{{criteria}}} ; container={{{container}}} ; nb_max={{{find_parameters.nb_max}}} ; visibility={{{find_parameters.visibility}}}")
        
        res = []
        criteria_parameters = find_parameters.get_criteria_parameters()
        
        # Search elements
        elements = container.get_list_holder_for(
                container.element.find_elements(by.by, by.value), 
                str(by) )
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"find_elements_by_container_by: by={{{by}}} ; criteria={{{criteria}}} ; container={{{container}}} ; nb_max={{{find_parameters.nb_max}}} ; visibility={{{find_parameters.visibility}}}       ->  {len(elements)} candidates found by selenium")
        
        # Filter on visibility
        for el in elements:
            if criteria is None or criteria.validate(el, find_context, criteria_parameters):
                if self.is_visible(el, find_context, find_parameters):
                    res.append(el)
                    
                    # Stop if max number of elements is reached
                    if find_parameters.nb_max_elements is not None and len(res) >= find_parameters.nb_max_elements:
                        break
                else:
                    if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"Element is {'not ' if find_parameters.visibility else ''}visible (element: [{self.get_element_description(el)}])")
                    
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"find_elements_by_container_by: by={{{by}}} ; criteria={{{criteria}}} ; container={{{container}}} ; nb_max={{{find_parameters.nb_max}}} ; visibility={{{find_parameters.visibility}}}       ->  {len(res)} elements")
        return res
    
    def find_parent_element(self, element, find_context, find_parameters):
        # Limit to root element
        root_element = None
        if isinstance(find_context,  WithRootContainerFindContext):
            root_element = find_context.root_container
        if root_element is not None and element.element == root_element.element:
            if find_parameters.raise_no_such_element:
                raise NoSuchElementException("Root element is reached.")
            else:
                return None
        
        # Search parent
        try:
            return self.find_element_by_container_by(SeleniumBy(By.XPATH, ".."), element, find_context, find_parameters)
        except TooManyElementsException as exc:
            raise TechnicalException(str(exc)) from exc
        except NoSuchElementException as exc:
            if find_parameters.raise_no_such_element:
                raise exc
            else:
                return None
        except (InvalidSelectorException | StaleElementReferenceException) as exc:
            if "HTMLDocument" in str(exc):
                # When HTMLDocument is reached, FirefoxDriver throws a StaleElementReferenceException, the other drivers an InvalidSelectorException
                if find_parameters.raise_no_such_element:
                    raise NoSuchElementException("HTMLDocument is reached.")
                else:
                    return None
            else:
                raise exc
    
    def find_children_elements(self, container, find_context, find_parameters):
        return self.find_elements_by_container_by(SeleniumBy(By.XPATH, "./*"), None, container, find_context, find_parameters)
    
    def _get_element_text(self, element, find_context, find_parameters):
        if find_parameters.visibility is not None and find_parameters.visibility:
            return self.__get_element_text_visible(element, find_context, find_parameters)
        else:
            return self.get_text_content(element)
    
    def __get_element_text_visible(self, element, find_context, find_parameters):
        res = None
        sub_find_parameters = find_parameters.with_analyze_time_spent(False)
        sub_find_parameters_1 = sub_find_parameters.with_nb_max_elements(1)
        web_element = self._get_web_element(element)
        
        # Case of input with value attribute
        if web_element.tag_name in ["input", "textarea"]:
            res = web_element.get_attribute("value")
        
        
        # Case of select with selected option attribute
        if res is None and web_element.tag_name == "select":
            options = web_element.find_elements(By.TAG_NAME, "option")
            for option in options:
                if option.is_selected:
                    res = option.text.strip()
                    break
        
        # Else, case with inner inputs or selects
        if (res is None
            and ( len(self.find_elements_by_container_by(SeleniumBy(By.TAG_NAME, "input"), None, element, find_context, sub_find_parameters_1)) > 0
                or len(self.find_elements_by_container_by(SeleniumBy(By.TAG_NAME, "select"), None, element, find_context, sub_find_parameters_1)) > 0
                or len(self.find_elements_by_container_by(SeleniumBy(By.TAG_NAME, "textarea"), None, element, find_context, sub_find_parameters_1)) > 0
                ) ):
            res_list = []
            
            # Add text of each child
            children = self.find_children(None, element, find_context, sub_find_parameters.with_nb_max_elements(None))
            for child in children:
                if self.is_visible(child, find_context, sub_find_parameters):
                    res_list.append(self.__get_element_text_visible(child, find_context, sub_find_parameters))
                else:
                    if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"Element is not visible (element: [{self.get_element_description(child)}])")
            
            res = "".join(res_list)
        
        # Else, return inner text
        if res is None:
            res = web_element.text.strip()
            res = html.unescape(res).strip()
        
        return res
    
    def get_text_content(self, element):
        """
        @param element An element
        @return Element's text content, including hidden text 
        """
        web_element = self._get_web_element(element)
        return web_element.get_attribute("textContent")
    
    def get_inner_html(self, element):
        """
        @param element An element
        @return Inner HTML of given element 
        """
        web_element = self._get_web_element(element)
        return web_element.get_attribute("innerHTML")
    
    def get_outer_html(self, element):
        """
        @param element An element
        @return HTML of given element 
        """
        web_element = self._get_web_element(element)
        return web_element.get_attribute("outerHTML")
    
    def is_displayed(self, element):
        web_element = self._get_web_element(element)
        res = web_element.is_displayed()
        
        # Check style has not "display: none"
        if res:
            value = web_element.get_attribute("style")
            res = ("display: none" not in value)
        
        return res
    
    def is_visible(self, element, find_context, find_parameters):
        res = None
        web_element = self._get_web_element(element)

        if find_parameters.visibility is None:
            return True
        
        # Case of a select option
        if web_element.tag_name == "option":
            try:
                parent = self.find_parent(element, find_context, find_parameters.with_analyze_time_spent(False).without_raise())
            except NoSuchElementException:
                parent = None
            
            if parent is not None:
                parent_element = self._get_web_element(parent)
                if parent_element.tag_name == "select":
                    selected_options = Select(parent_element).all_selected_options
                    res = (find_parameters.visibility == (web_element in selected_options))
        
        # Else
        if res is None:
            res = (find_parameters.visibility == self.is_displayed(element))
        
        return res
    
    def move_mouse_over(self, element):
        web_element = self._get_web_element(element)
        ActionChains(self.internal_driver).move_to_element(web_element).perform()
    
    def make_screenshot(self, destination_path, context_description):
        done = False
        try_counter = 0
        while not done:
            try_counter += 1
            try:
                # Take screenshot
                super().make_screenshot(destination_path, context_description)
                done = True
            except FunctionalException as exc:
                if try_counter < 3:
                    logger.warning(f"Unable to make screenshot ; retry (error: {str(exc)})")
                else:
                    logger.error(f"Unable to make screenshot ; stop retries (error: {str(exc)})")
                    return
            except Exception as exc:
                logger.error(f"Unable to make screenshot (error: {str(exc)})")
                return
    
    def get_screenshot_file(self, screenshot_path, element=None, margin_around_element=0):
        # Implementation is using: https://pypi.org/project/Selenium-Screenshot/
        self.driver._path_manager.makedirs(screenshot_path)
        dir_path, filename = os.path.split(screenshot_path)
        ss = Screenshot.Screenshot()
        
        if element is None:
            # Take screenshot
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace("Getting screenshot file...")
            res = ss.full_screenshot(self.internal_driver, save_path=dir_path , image_name=filename)
        else:
            # Take screenshot
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace("Getting screenshot file of element...")
            res = ss.get_element(self.internal_driver, self._get_web_element(element), save_path=dir_path , image_name=filename)
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace("Got screenshot file '{}'", res.getPath())
        return res
    
    def make_page_source_backup(self, destination_path, context_description):
        """
        Make a backup of current page source.
        @param destination_path Destination path
        @param context_description Context description that will be inserted in file names
        """
        # Create file name
        date_str = DateTime.now().strftime('%Y-%m-%d_%H-%M-%S.%f')
        file_name = f"{date_str}-{context_description}-page_source.html"
        path = os.path.join(destination_path, file_name)
        
        page_source = None
        try_counter = 0
        while page_source is None:
            try_counter += 1
            try:
                # Take sources
                page_source = self.get_page_source()
            except FunctionalException as exc:
                if try_counter < 3:
                    logger.warning(f"Unable to make browser page source backup ; retry (error: {str(exc)})")
                else:
                    logger.error(f"Unable to make browser page source backup ; stop retries (error: {str(exc)})")
                    return
            except Exception as exc:
                logger.error(f"Unable to make browser page source backup (error: {str(exc)})")
                return
        
        # Create file
        self.driver._path_manager.makedirs(path)
        with open(path, 'wt') as fout:
            fout.write(page_source)
    
    def get_page_source(self):
        """
        Get browser page source.
        @return File containing screenshot
        """
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace("Getting page source...")
        
        return self.internal_driver.page_source
    
    def get_element_description(self, element, with_text=True):
        if element.is_web_element:
            res_list = []
            web_element = self._get_web_element(element)
            id_ = web_element.get_attribute("id")
            if id_ is None or len(id_) == 0:
                id_ = web_element.id
            
            res_list.append(f"id={id_}", )
            res_list.append(f"tag={web_element.tag_name}")
            res_list.append(f"coordinates={self.get_element_coordinates(element)}")
            res_list.append(f"size={self.get_element_size(element)}")
            if with_text:
                res_list.append(f"text=[{Tools.truncate_text(self.get_element_text(element), 1000)}]")
            res_list.append(f"html=[{Tools.truncate_text(self.get_outer_html(element), 1000)}]")
            
            return " ; ".join(res_list)
        else:
            return element.description
    
    def scroll_into_view(self, element, sleep_after_scroll_ms):
        if self.internal_driver is None:
            raise TechnicalException("Driver is None")
            
        self.internal_driver.execute_script("arguments[0].scrollIntoView(True)", self._get_web_element(element))
        time.sleep(sleep_after_scroll_ms/1000)
    
    def scroll_to(self, x, y, sleep_after_scroll_ms):
        if self.internal_driver is None:
            raise TechnicalException("Driver is None")
            
        self.internal_driver.execute_script(f"window.scrollTo({x}, {y})")
        time.sleep(sleep_after_scroll_ms/1000)
    
    def scroll_by(self, x, y, sleep_after_scroll_ms):
        if self.internal_driver is None:
            raise TechnicalException("Driver is None")
            
        self.internal_driver.execute_script(f"window.scrollBy({x}, {y})")
        time.sleep(sleep_after_scroll_ms/1000)
    
    def generate_xpath(self, web_element):
        """
        @param element Element
        @return Generated xpath of element
        """
        return self.__generate_xpath(web_element, "")

    def __generate_xpath(self, child_element:WebElement, current):
        child_tag = child_element.tag_name
        if child_tag == "html":
            return "/html[1]"+current
        
        parent_element = child_element.find_element(By.XPATH, "..")
        children_elements = parent_element.find_elements(By.XPATH, "*")
        count = 0
        for children_element in children_elements:
            children_element_tag = children_element.tag_name
            if child_tag == children_element_tag:
                count += 1
            
            if child_element == children_element:
                return self.__generate_xpath(parent_element, f"/{child_tag}[{count}]{current}")
        
        return None
    
    def refresh_browser(self):
        """
        Refresh browser
        """
        self.internal_driver.refresh()
    
    #TODO move in windows in order to adapt behavior on window type
    def wait_until_window_is_loaded(self, timeout_seconds):
        if not self.is_activated_wait_until_window_is_loaded():
            logger.info("Skipped wait until window is loaded (deactivated)")
            return
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug("Waiting until window is loaded...")
        
        class WinRedo(Redo):
            def __init__(self, driver):
                super().__init__("window is loaded")
                self.__driver = driver
                
            def _process(self):
                res = False
                
                try:
                    if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        msg_list = ["Waiting until window is loaded: "]

                    # Verify HTML document
                    if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        msg_list.append("HTML document")
                        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                            logger.trace(f"{''.join(msg_list)}...")
                    
                    res = (self.internal_driver.execute_script("return document.readyState") == "complete")
                    if not res and logger.isTraceEnabled():
                        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                            logger.trace(f"{''.join(msg_list)} is KO")
                    
                    # Verify jQuery
                    if res and self.__wait_jquery:
                        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                            msg_list.append(" is OK  jQuery")
                            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                                logger.trace(f"{''.join(msg_list)}...")
                        
                        res = Converter.to_boolean(self.internal_driver.execute_script("return window.jQuery == undefined || jQuery.active === 0"))
                        if not res and logger.isTraceEnabled():
                            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                                logger.trace(f"{''.join(msg_list)} is KO")
                    
                    
                    # Verify angular
                    if res and self.__wait_angular:
                        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                            msg_list.append(" is OK  angular")
                            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                                logger.trace(f"{''.join(msg_list)}...")
                        
                        res = Converter.to_boolean(self.internal_driver.execute_script("return window.angular == undefined || angular.element(document).injector() !== undefined && angular.element(document).injector().get('$http').pendingRequests.length === 0"))
                        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                            logger.trace(f"{''.join(msg_list)} is {'OK' if res else 'KO'}")
                    
                except UnexpectedAlertPresentException as exc:
                    raise exc
                except WebDriverException as exc:
                    if str(exc).startswith("[JavaScript Error: "):
                        # FIXME: find a better workaround for this JavaScript error
                        logger.warning(f"Window is considered loaded when a JavaScript error occurs ; JavaScript error: {str(exc)}")
                        res = True 
                    else:
                        logger.warning(f"Retry because of unexpected exception of type '{Typing.get_object_class_fullname(exc)}' during Javascript call: {str(exc)}")
                        res = False
                
                return res
            
            def _process_interrupt(self, thread):
                # This call is needed, so that the blocking executeScript is stopped.
                # Otherwise, the process method continue until it finishes normally.
                
                # Following step is not working anymore, it directly throws exception UnsupportedOperationException
#                thread.stop(new InterruptedException())
                thread.stop()
            
        redo = WinRedo()
        redo.redo_while(False)
        redo.with_process_timeout(10)
        redo.with_nb_successive_failure_before_error(3)
        if timeout_seconds is not None:
            redo.with_timeout(timeout_seconds)

        try:
            redo.execute()
        except UnexpectedAlertPresentException:
            # Nothing  next sentence should manage this alert
            pass
    
    def add_cookie(self, cookie):
        """
        @param cookie Cookie
        """
        self.internal_driver.add_cookie(cookie)
    
    def disactivate_js_on_before_unload(self):
        self.internal_driver.execute_script("window.onbeforeunload = function(e){}")
    
    def click_on_element_with_selenium(self, element):
        """
        @param element Element
        """
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"clicking with selenium on element [{self.get_element_description(element)}] (source: [{self.get_outer_html(element)}])")
        self._get_web_element(element).click()
    
    def click_on_element_with_javascript(self, element):
        """
        @param element Element
        """
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"clicking with javascript on element [{self.get_element_description(element)}] (source: [{self.get_outer_html(element)}])")
        
        self.internal_driver.execute_script("arguments[0].click()", self._get_web_element(element))
    
    def write_to_element_all_keys(self, element, value):
        """
        @param element Element
        @param value Value
        """
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"writing text '{value}' in element [{self.get_element_description(element)}]")
        
        # Write text
        self.send_keys(element, value)
        self.wait_until_window_is_loaded()
    
    def write_to_element_key_by_key(self, element, value):
        """
        @param element Element
        @param value Value
        """
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"writing (key by key) text '{value}' in element [{self.get_element_description(element)}]")

        # Send each text character
        for c in value:
            self.send_keys(element, c)
            self.wait_until_window_is_loaded()
        
    def send_key_multiple_times_all_keys(self, element, key:Keys, nb_times):
        """
        @param element Element
        @param key Key
        @param nb_times Number times
        """
        keys_list = []
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Sending {nb_times} {key} to element [{self.get_element_description(element)}]")

        # Add as many backspace as existing text length
        for _ in range(nb_times):
            # Call backspace character
            keys_list.append(key)
        
        
        # Send keys
        self.send_keys(element, "".join(keys_list))
        self.wait_until_window_is_loaded()
    
    def send_key_multiple_times_key_by_key(self, element, key:Keys, nb_times):
        """
        @param element Element
        @param key Key
        @param nb_times Number times
        """
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Sending (key by key) {nb_times} {key} to element [{self.get_element_description(element)}]")

        # Send as many backspace as existing text length
        for _ in range(nb_times):
            self.send_keys(element, key)
            self.wait_until_window_is_loaded()
    
    def send_keys(self, element, keys):
        """
        @param element Web element
        @param keys Keys to send (as string or Keys value)
        """
        self._get_web_element(element).send_keys(keys)
    
    def clear_element(self, element):
        self._get_web_element(element).clear()
    
    def focus(self):
        self.internal_driver.execute_script("window.focus()")
    
    def maximize(self):
        self.internal_driver.maximize_window()
        self.wait_until_window_is_loaded()
    
    def get_cookies(self):
        """
        @return Cookies
        """
        return self.internal_driver.get_cookies()
    
    def open_new_window(self):
        self.internal_driver.execute_script("window.open()")
    
    def is_stale(self, element):
        """
        @param element Web element
        @return If element is stale
        """
        return EC.staleness_of(self._get_web_element(element)).apply(self.internal_driver)
    
    def getWindowHandles(self):
        """
        @return List of driver handles
        """
        try:
            return self.internal_driver.window_handles
        except SessionNotCreatedException:
            return []
    
    def get_element_coordinates(self, element):
        raise NotImplementedError
    
    def get_element_location(self, element):
        return self._get_web_element(element).location
    
    def get_element_rectangle(self, element):
        return self._get_web_element(element).rect
    
    def get_element_size(self, element):
        return self._get_web_element(element).size
    
    def _get_web_element(self, element_holder):
        if element_holder.is_web_element:
            return element_holder.element
        else:
            raise TechnicalException(f"Holded element is of type '{Typing.get_object_class_fullname(element_holder.element)}' rather than WebElement")
    
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from builtins import super
from holado.common.context.context import Context
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_scripting.common.tools.variable_manager import VariableManager
from holado.common.context.session_context import SessionContext
from holado_python.common.tools.datetime import DateTime

class FeatureContext(Context):
    def __init__(self, feature):
        super().__init__("Feature")
        
        self.__feature = feature
        
        self.__start_date = DateTime.now()
        self.__end_date = None
        self.__scenarios = []
    
    def __str__(self):
        return f"{{FeatureContext({id(self)}):{self.feature.name}}}"
     
    @property
    def feature(self):
        return self.__feature
    
    @property
    def start_datetime(self):
        return self.__start_date
    
    @property
    def end_datetime(self):
        return self.__end_date
    
    @property
    def duration(self):
        if self.__end_date is not None:
            return (self.__end_date - self.__start_date).total_seconds()
        else:
            return None
    
    @property
    def has_scenario(self):
        return len(self.__scenarios) > 0
    
    @property
    def current_scenario(self):
        if not self.has_scenario:
            raise TechnicalException("Feature has no scenario")
        return self.__scenarios[-1]
    
    def add_scenario(self, scenario_context):
        self.__scenarios.append(scenario_context)
    
    def end(self):
        self.__end_date = DateTime.now()
        
    def has_variable_manager(self):
        return self.has_object("variable_manager")
        
    def get_variable_manager(self) -> VariableManager:
        if not self.has_variable_manager():
            manager = VariableManager(SessionContext.instance().multitask_manager.get_thread_context().get_variable_manager())
            self.set_object("variable_manager", manager)
            file_path = SessionContext.instance().report_manager.current_feature_report.get_path("logs", "variable_update.log") if SessionContext.instance().with_session_path else None
            manager.initialize(SessionContext.instance().dynamic_text_manager, SessionContext.instance().unique_value_manager,
                               variable_update_log_file_path=file_path)
        return self.get_object("variable_manager")
    
        




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado.common.context.session_context import SessionContext
from holado.common.handlers.object import Object
import abc
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_python.standard_library.typing import Typing
from holado_yaml.yaml.enums import UpdateType
import os
from holado_python.common.iterables import remove_all
from holado.common.handlers.undefined import not_applicable, undefined_value
from io import StringIO
from holado_core.common.tools.tools import Tools


logger = logging.getLogger(__name__)


class YAMLClient(Object):
    """
    Client for actions on YAML files.
    """
    __metaclass__ = abc.ABCMeta
    
    @classmethod
    def _get_path_manager(cls):
        return SessionContext.instance().path_manager
    
    def __init__(self, name=None):
        super().__init__(name)
    
    def load_file(self, file_path):
        with open(file_path, 'r') as file:
            res = self.load_io_file(file)
            
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Load YAML file '{file_path}' => [{Typing.get_object_class_fullname(res)}] {res}")
        return res
    
    def load_string(self, text):
        with StringIO(text) as file:
            res = self.load_io_file(file)
            
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Load YAML string [{text}] => [{Typing.get_object_class_fullname(res)}] {res}")
        return res
    
    def load_io_file(self, file_like_object):
        raise NotImplementedError()
    
    def load_multiple_documents_file(self, file_path):
        with open(file_path, 'r') as file:
            res = self.load_multiple_documents_io_file(file)
        return res
        
    def load_multiple_documents_string(self, text):
        with StringIO(text) as file:
            res = self.load_multiple_documents_io_file(file)
        return res
        
    def load_multiple_documents_io_file(self, file_like_object):
            raise NotImplementedError()
        
    def save_in_file(self, file_path, data, mode=None, user=None, group=None, **kwargs):
        self._get_path_manager().makefile(file_path, mode=mode, user=user, group=group)
        
        with open(file_path, 'w') as file:
            self.save_in_io_file(file, data, **kwargs)
    
    def save_in_string(self, data, **kwargs):
        with StringIO() as file:
            self.save_in_io_file(file, data, **kwargs)
            res = file.getvalue()
        return res
    
    def save_in_io_file(self, file_like_object, data, **kwargs):
        raise NotImplementedError()
    
    def update_file(self, file_path, data, update_type=UpdateType.Default, with_backup=True, backup_extension='.bak', mode=None, user=None, group=None, **kwargs):
        # If file doesn't exist, create an empty file
        self._get_path_manager().makefile(file_path, mode=mode, user=user, group=group)
        
        # Manage backup
        # Note: It is done after potential empty file creation, so that restore will leave the file empty or remove it  
        if with_backup:
            if backup_extension is None:
                backup_extension = '.bak'
            backup_path = file_path + backup_extension
            if not self._get_path_manager().check_file_exists(backup_path, raise_exception=False):
                self._get_path_manager().copy_path_recursively(file_path, backup_path, user=user, group=group)
                self._get_path_manager().check_file_exists(backup_path, raise_exception=True)
        
        # Update file
        dst_data = self.load_file(file_path)
        if dst_data is None:
            dst_data = data
        else:
            self.update_data(dst_data, data, update_type)
        self.save_in_file(file_path, dst_data, **kwargs)        # Note: as file already exists, it is not needed to pass parameters mode, user and group
    
    def update_string(self, text, data, update_type=UpdateType.Default, **kwargs):
        dst_data = self.load_string(text)
        if dst_data is None:
            dst_data = data
        else:
            self.update_data(dst_data, data, update_type)
        res = self.save_in_string(dst_data, **kwargs)
        return res
    
    def update_data(self, dst, src, update_type=UpdateType.Default):
        if dst is None:
            raise TechnicalException("Destination data cannot be None")
        if isinstance(src, str):
            src = self.load_string(src)
        
        self._update_data_object(dst, src, update_type, level=0)
    
    def _update_data_object(self, dst, src, update_type, level):
        if isinstance(src, dict):
            self._update_data_dict(dst, src, update_type, level)
        elif isinstance(src, list):
            self._update_data_list(dst, src, update_type, level)
        else:
            raise TechnicalException(f"Unmanaged update of object of type {Typing.get_object_class_fullname(dst)} with data of type {Typing.get_object_class_fullname(src)}")

    def __is_list_or_dict(self, value):
        res = isinstance(value, list) or isinstance(value, dict)
        if Tools.do_log(logger, logging.TRACE):
            logger.trace(f"__is_list_or_dict([{type(value)}] {value}) => {res}")
        return res

    def _update_data_dict(self, dst, src, update_type, level):
        if Tools.do_log(logger, logging.TRACE):
            logger.trace(f"_update_data_dict({dst=}, {src=}, {update_type=}, {level=})")
        if update_type in [UpdateType.Default, UpdateType.AddOrUpdate]:
            if Tools.do_log(logger, logging.TRACE):
                logger.trace(f"_update_data_dict({dst=}, {src=}, {update_type=}, {level=}) -> AddOrUpdate")
            for key, value in src.items():
                if key not in dst or not self.__is_list_or_dict(value):
                    dst[key] = value
                else:
                    self._update_data_object(dst[key], value, update_type, level + 1)
        elif update_type == UpdateType.Delete:
            if Tools.do_log(logger, logging.TRACE):
                logger.trace(f"_update_data_dict({dst=}, {src=}, {update_type=}, {level=}) -> Delete")
            for key, value in src.items():
                if key in dst:
                    if self.__is_list_or_dict(value):
                        self._update_data_object(dst[key], value, update_type, level + 1)
                        if not dst[key]:
                            del dst[key]
                    else:
                        del dst[key]
        elif update_type == UpdateType.Replace:
            if Tools.do_log(logger, logging.TRACE):
                logger.trace(f"_update_data_dict({dst=}, {src=}, {update_type=}, {level=}) -> Replace")
            has_inner_iterable = any(self.__is_list_or_dict(value) for value in src.values())
            if not has_inner_iterable:
                if Tools.do_log(logger, logging.TRACE):
                    logger.trace(f"_update_data_dict({dst=}, {src=}, {update_type=}, {level=}) -> Replace -> clear")
                dst.clear()

            for key, value in src.items():
                if key in dst and self.__is_list_or_dict(value):
                    self._update_data_object(dst[key], value, update_type, level + 1)
                else:
                    dst[key] = value
        elif update_type == UpdateType.ReplaceElements:
            if Tools.do_log(logger, logging.TRACE):
                logger.trace(f"_update_data_dict({dst=}, {src=}, {update_type=}, {level=}) -> ReplaceElements")
            for key, value in src.items():
                if key in dst and self.__is_list_or_dict(value):
                    self._update_data_object(dst[key], value, update_type, level + 1)
                else:
                    dst[key] = value
        else:
            raise TechnicalException(f"Unmanaged update type {update_type}")
    
    def _update_data_list(self, dst, src, update_type, level):
        if Tools.do_log(logger, logging.TRACE):
            logger.trace(f"_update_data_list({dst=}, {src=}, {update_type=}, {level=})")
        # If src is a list of primitive types, replace update_type Default by Replace
        if update_type == UpdateType.Default and not any(self.__is_list_or_dict(x) for x in src):
            update_type = UpdateType.Replace

        if update_type in [UpdateType.Default, UpdateType.AddOrUpdate]:
            if Tools.do_log(logger, logging.TRACE):
                logger.trace(f"_update_data_list({dst=}, {src=}, {update_type=}, {level=}) -> AddOrUpdate")
            for value in src:
                dst.append(value)
        elif update_type == UpdateType.Delete:
            if Tools.do_log(logger, logging.TRACE):
                logger.trace(f"_update_data_list({dst=}, {src=}, {update_type=}, {level=}) -> Delete")
            for value in src:
                remove_all(dst, value)
        elif update_type == UpdateType.Replace:
            if Tools.do_log(logger, logging.TRACE):
                logger.trace(f"_update_data_list({dst=}, {src=}, {update_type=}, {level=}) -> Replace")
            has_inner_iterable = any(self.__is_list_or_dict(value) for value in src)
            if not has_inner_iterable:
                if Tools.do_log(logger, logging.TRACE):
                    logger.trace(f"_update_data_list({dst=}, {src=}, {update_type=}, {level=}) -> Replace -> clear")
                dst.clear()

            for value in src:
                dst.append(value)
        elif update_type == UpdateType.ReplaceElements:
            if Tools.do_log(logger, logging.TRACE):
                logger.trace(f"_update_data_list({dst=}, {src=}, {update_type=}, {level=}) -> ReplaceElements")
            for index, value in enumerate(src):
                if value is not_applicable:
                    continue
                if index < len(dst):
                    if self.__is_list_or_dict(value):
                        self._update_data_object(dst[index], value, update_type, level + 1)
                    else:
                        dst[index] = value
                else:
                    while len(dst) < index:
                        dst.append(undefined_value)
                    dst.append(value)
        else:
            raise TechnicalException(f"Unmanaged update type {update_type}")
    
    def restore_file(self, file_path, backup_extension='.bak', remove_empty_file=False):
        if backup_extension is None:
            backup_extension = '.bak'
        backup_path = file_path + backup_extension
        if self._get_path_manager().check_file_exists(backup_path, raise_exception=False):
            # Replace file by backup
            self._get_path_manager().rename(backup_path, file_path, raise_if_exists=False)
            
            # Manage remove of empty file
            if remove_empty_file and os.path.getsize(file_path) == 0:
                self._get_path_manager().remove_path(file_path)






import copy
#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging

from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_rest.api.rest.rest_client import RestClient
import socket
import urllib.parse
import re
from holado_rest.api.rest.rest_manager import RestManager
from holado.common.handlers.undefined import undefined_argument, undefined_value
import os
from holado_core.common.tools.converters.converter import Converter
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.handlers.wait import WaitFuncResult

logger = logging.getLogger(__name__)


class KeycloakClient(RestClient):
    """
    Helper class to interact with the Keycloak REST API.
    It inherits from RestClient all methods of a REST client, thus any endpoint can be called.
    Some methods are added for commonly used endpoints, and illustrates how use this client for other endpoints.
    """

    @classmethod
    def new_client(cls, use_localhost=undefined_argument, **kwargs):
        if 'name' not in kwargs:
            kwargs['name'] = None
        if 'url' not in kwargs:
            if use_localhost is undefined_argument:
                env_use = os.getenv("HOLADO_USE_LOCALHOST", False)
                use_localhost = Converter.is_boolean(env_use) and Converter.to_boolean(env_use)
            
            scheme = kwargs.get('scheme', 'http')
            host = kwargs.get('host', 'localhost' if use_localhost else undefined_value)
            if host is undefined_value:
                raise FunctionalException("Host or url must be defined")
            port = kwargs.get('port', None)

            if port is None:
                url = f"{scheme}://{host}"
            else:
                url = f"{scheme}://{host}:{port}"
            kwargs['url'] = url

        # Force content-type to application/x-www-form-urlencoded

        manager = RestManager(default_client_class=cls)
        res = manager.new_client(**kwargs)
        
        return res
    
    
    def __init__(self, name, url, headers=None):
        super().__init__(name, url, headers)

    def get_token(self, **kwargs):
        if 'username' in kwargs and 'password' in kwargs:
            return self.get_token_by_grant_type_password(**kwargs)
        elif 'refresh_token' in kwargs:
            return self.get_token_by_grant_type_refresh_token(**kwargs)
        elif 'client_id' in kwargs and 'client_secret' in kwargs:
            return self.get_token_by_grant_type_client_credentials(**kwargs)
        else:
            raise FunctionalException(
                f"Unable to get token with given parameters.\nAllowed methods and associated parameters: password -> ('username', 'password'), refresh -> ('refresh_token'), client credentials -> ('client_id', 'client_secret')\nReceived parameters: {kwargs}")

    def get_token_by_grant_type_password(self, realm_name, username, password, client_id=None, client_secret=None):
        data = {
            'grant_type': 'password',
            'username': username,
            'password': password,
        }
        if client_id:
            data['client_id'] = client_id
        if client_secret:
            data['client_secret'] = client_secret

        return self.post(f"realms/{realm_name}/protocol/openid-connect/token", data=data, headers={"Content-Type": "application/x-www-form-urlencoded"})

    def get_token_by_grant_type_refresh_token(self, realm_name, refresh_token, client_id=None, client_secret=None):
        data = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token,
        }
        if client_id:
            data['client_id'] = client_id
        if client_secret:
            data['client_secret'] = client_secret

        return self.post(f"realms/{realm_name}/protocol/openid-connect/token", data=data, headers={"Content-Type": "application/x-www-form-urlencoded"})

    def get_token_by_grant_type_client_credentials(self, realm_name, client_id=None, client_secret=None):
        data = {
            'grant_type': 'client_credentials',
        }
        if client_id:
            data['client_id'] = client_id
        if client_secret:
            data['client_secret'] = client_secret

        return self.post(f"realms/{realm_name}/protocol/openid-connect/token", data=data, headers={"Content-Type": "application/x-www-form-urlencoded"})

    def logout(self, **kwargs):
        """ Logout from Keycloak
        Note: only method with refresh_token seems to work, thus only this method is currently managed.
        """
        if 'refresh_token' in kwargs:
            return self.logout_by_refresh_token(**kwargs)
        # NOTE: As described below, method by token should work but doesn't, that's why it is commented
        # elif 'token' in params:
        #     client.logout_by_token(**params)
        else:
            raise FunctionalException(f"Unable to logout with given parameters.\nAllowed methods and associated parameters: refresh token -> ('refresh_token')\nReceived parameters: {kwargs}")

    # NOTE: This method doesn't work whereas it should according to documentation. Use instead logout_by_refresh_token.
    # def logout_by_token(self, realm_name, token, client_id=None, client_secret=None):
    #     data = {
    #         'id_token_hint': token,
    #     }
    #     if client:
    #         data['client_id'] = client[0]
    #         data['client_secret'] = client[1]
    #
    #     return self.post(f"realms/{realm_name}/protocol/openid-connect/logout", json=data)

    def logout_by_refresh_token(self, realm_name, refresh_token, client_id=None, client_secret=None):
        data = {
            'refresh_token': refresh_token,
        }
        if client_id:
            data['client_id'] = client_id
        if client_secret:
            data['client_secret'] = client_secret

        return self.post(f"realms/{realm_name}/protocol/openid-connect/logout", data=data, headers={"Content-Type": "application/x-www-form-urlencoded"})

    def get_user_info(self, realm_name, username, client_id=None, client_secret=None, field_name=None, field_default_value=None, do_raise_exception=True):
        data = {}
        if client_id:
            data['client_id'] = client_id
        if client_secret:
            data['client_secret'] = client_secret

        path_parameters = {
            'exact': True,
            'username': username,
        }

        if data:
            result = self.get(f"admin/realms/{realm_name}/users", data=data, path_parameters=path_parameters, headers={"Content-Type": "application/x-www-form-urlencoded"})
        else:
            result = self.get(f"admin/realms/{realm_name}/users", path_parameters=path_parameters)

        if len(result) == 0:
            if do_raise_exception:
                raise FunctionalException(f"Unable to find information on user '{username}'.")
            else:
                return None
        elif len(result) > 1:
            raise TechnicalException(f"Found {len(result)} users for user '{username}': {result}")
        res = result[0]

        if field_name is not None:
            res = res.get(field_name, field_default_value)
        return res

    def get_user_required_actions(self, realm_name, username, client_id=None, client_secret=None):
        return self.get_user_info(realm_name, username, client_id=client_id, client_secret=client_secret, field_name="requiredActions", field_default_value=[])

    def get_user_id(self, realm_name, username, client_id=None, client_secret=None):
        user = self.get_user_info(realm_name, username, client_id=client_id, client_secret=client_secret)
        return user['id']

    def verify_user_exists(self, realm_name, username, do_raise_exception=True):
        user_info = self.get_user_info(realm_name, username, do_raise_exception=do_raise_exception)
        return user_info is not None

    def create_user(self, realm_name, username, **user_data):
        data = copy.deepcopy(user_data)
        data['username'] = username

        return self.post(f"admin/realms/{realm_name}/users", json=data)

    def set_user_password(self, realm_name, username, password, client_id=None, client_secret=None):
        user_id = self.get_user_id(realm_name, username, client_id=client_id, client_secret=client_secret)

        data = {
            'type': 'password',
            'temporary': False,
            'value': password,
        }
        if client_id:
            data['client_id'] = client_id
        if client_secret:
            data['client_secret'] = client_secret

        # return self.put(f"admin/realms/{realm_name}/users/{user_id}/reset-password", data=data, headers={"Content-Type": "application/x-www-form-urlencoded"})
        return self.put(f"admin/realms/{realm_name}/users/{user_id}/reset-password", json=data)

    def get_client_info(self, realm_name, client_id, do_raise_exception=True):
        path_parameters = {
            'clientId': client_id,
        }

        result = self.get(f"admin/realms/{realm_name}/clients", path_parameters=path_parameters)

        if len(result) == 0:
            if do_raise_exception:
                raise FunctionalException(f"Unable to find information on client '{client_id}'.")
            else:
                return None
        elif len(result) > 1:
            raise TechnicalException(f"Found {len(result)} clients for client '{client_id}': {result}")
        return result[0]

    def verify_client_exists(self, realm_name, client_id, do_raise_exception=True):
        client_info = self.get_client_info(realm_name, client_id, do_raise_exception=do_raise_exception)
        return client_info is not None

    def create_client(self, realm_name, client_id, **client_data):
        data = copy.deepcopy(client_data)
        data['clientId'] = client_id

        # return self.post(f"admin/realms/{realm_name}/clients", data=data, headers={"Content-Type": "application/x-www-form-urlencoded"})
        return self.post(f"admin/realms/{realm_name}/clients", json=data)

    def verify_user_has_required_action(self, realm_name, username, action, check_has=True, do_raise_exception=True):
        required_actions = self.get_user_required_actions(realm_name, username)
        if check_has:
            res = action in required_actions
        else:
            res = action not in required_actions

        if not res and do_raise_exception:
            raise FunctionalException(f"User '{username}' has{' not' if check_has else ''} required action '{action}' (required actions: {required_actions}).")
        return res

    def process_automatically_user_required_actions(self, realm_name, username):
        """
        Process automatically some user actions.
        :param realm_name: realm name
        :param username: username
        """
        required_actions = self.get_user_required_actions(realm_name, username)
        # Do a loop as workflows can create new actions
        while required_actions:
            # Process actions
            for action in required_actions:
                if action == "VERIFY_EMAIL":
                    self.process_action_user_verify_email(realm_name, username)
                else:
                    raise FunctionalException(f"Unmanaged user action '{action}'")

            required_actions = self.get_user_required_actions(realm_name, username)

    def process_action_user_verify_email(self, realm_name, username):
        user_id = self.get_user_id(realm_name, username)

        data = {"emailVerified": True}
        self.put(f"admin/realms/{realm_name}/users/{user_id}", json=data)

        # As user is forced with emailVerified=true, action was not really performed
        if self.verify_user_has_required_action(realm_name, username, action="VERIFY_EMAIL", do_raise_exception=False):
            self.remove_user_required_action(realm_name, username, action="VERIFY_EMAIL", user_id=user_id)
            self.verify_user_has_required_action(realm_name, username, action="VERIFY_EMAIL", check_has=False, do_raise_exception=True)

    def remove_user_required_action(self, realm_name, username, action, user_id=None):
        if user_id is None:
            user_id = self.get_user_id(realm_name, username)

        required_actions = self.get_user_required_actions(realm_name, username)
        if action not in required_actions:
            raise FunctionalException(f"Action '{action}' is not in required actions: {required_actions}.")

        required_actions.remove(action)
        data = {"requiredActions": required_actions}
        self.put(f"admin/realms/{realm_name}/users/{user_id}", json=data)

    def ensure_user_is_fully_set_up(self, realm_name, username):
        user_info = self.get_user_info(realm_name, username)

        # Ensure email is verified
        if not user_info["emailVerified"]:
            self.process_action_user_verify_email(realm_name, username)





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging


# logger = logging.getLogger(__name__)


class LogConfig(object):
    TLogger = None
    TManager = None
    config_file_path = None
    default_level = logging.INFO
    log_time_in_utc = True
    log_on_console=False
    log_in_file=True
    __is_configured = False
    
    @classmethod
    def configure(cls, use_holado_logger=True, config_file_path=None, log_level=None, log_time_in_utc=None, log_on_console=False, log_in_file=True):
        if cls.__is_configured:
            logging.warning(f"Logging was already configured, it is not possible to configure it twice. This new configuration is skipped.")
            return
        
        # Use holado loggers
        if use_holado_logger:
            cls.__set_holado_loggers()
        
        # HolAdo needs at least to add logging level TRACE and PRINT
        cls.add_logging_level_trace()
        cls.add_logging_level_print()
        
        cls.config_file_path = config_file_path
        if config_file_path:
            import configparser
            config = configparser.ConfigParser()
            config.read(config_file_path)
            log_level = config.get("holado", "level", fallback=log_level)
            log_time_in_utc = config.getboolean("holado", "log_time_in_utc", fallback=log_time_in_utc)
            log_on_console = config.get("holado", "log_on_console", fallback=log_on_console)
            log_in_file = config.get("holado", "log_in_file", fallback=log_in_file)
        
        if log_level:
            if isinstance(log_level, str):
                log_level = logging._nameToLevel[log_level]
            cls.default_level = log_level
        if log_time_in_utc is not None:
            cls.log_time_in_utc = log_time_in_utc
        if log_on_console:
            if isinstance(log_on_console, str):
                log_on_console = True if log_on_console == "True" else False
            cls.log_on_console = log_on_console
        if log_in_file:
            if isinstance(log_in_file, str):
                log_in_file = True if log_in_file == "True" else False
            cls.log_in_file = log_in_file
        
        # Change log time format if needed
        if cls.log_time_in_utc:
            import time
            logging.Formatter.converter = time.gmtime
        
        cls.__is_configured = True
    
    @classmethod
    def __set_holado_loggers(cls):
        from holado_logging.common.logging.holado_logger import HALogger
        
        # Configure loggers to use
        HALogger.default_message_size_limit = 10000
        cls.TLogger = HALogger
        cls.TManager = logging.Manager
        
        # Set loggers in logging
        from holado_logging.common.logging.holado_logger import HARootLogger
        logging.root = HARootLogger(cls.default_level)
        logging.Logger.root = logging.root
        logging.Logger.manager = cls.TManager(cls.TLogger.root)
        
        logging.setLoggerClass(cls.TLogger)
        
    
    @classmethod
    def add_logging_level_print(cls):
        if not cls.has_logging_level("PRINT"):
            cls.add_logging_level("PRINT", 45, None)
    
    @classmethod
    def add_logging_level_trace(cls):
        if not cls.has_logging_level("TRACE"):
            cls.add_logging_level("TRACE", 5, None)
        
    
    @classmethod
    def has_logging_level(cls, levelName):
        return hasattr(logging, levelName)
    
    @classmethod
    def add_logging_level(cls, levelName, levelNum, methodName=None):
        """
        This method was implemented and shared by the author of library haggis (https://haggis.readthedocs.io).
        
        Comprehensively adds a new logging level to the `logging` module and the
        currently configured logging class.
    
        `levelName` becomes an attribute of the `logging` module with the value
        `levelNum`. `methodName` becomes a convenience method for both `logging`
        itself and the class returned by `logging.getLoggerClass()` (usually just
        `logging.Logger`). If `methodName` is not specified, `levelName.lower()` is
        used.
    
        To avoid accidental clobberings of existing attributes, this method will
        raise an `AttributeError` if the level name is already an attribute of the
        `logging` module or if the method name is already present 
    
        Example
        -------
        >>> addLoggingLevel('TRACE', logging.DEBUG - 5)
        >>> logging.getLogger(__name__).setLevel("TRACE")
        >>> logging.getLogger(__name__).trace('that worked')
        >>> logging.trace('so did this')
        >>> logging.TRACE
        5
    
        """
        if not methodName:
            methodName = levelName.lower()
    
        if hasattr(logging, levelName):
            raise AttributeError('{} already defined in logging module'.format(levelName))
        if hasattr(logging, methodName):
            raise AttributeError('{} already defined in logging module'.format(methodName))
        if hasattr(logging.getLoggerClass(), methodName):
            raise AttributeError('{} already defined in logger class'.format(methodName))
    
        # This method was inspired by the answers to Stack Overflow post
        # http://stackoverflow.com/q/2183233/2988730, especially
        # http://stackoverflow.com/a/13638084/2988730
        def logForLevel(self, message, *args, **kwargs):
            if self.isEnabledFor(levelNum):
                self._log(levelNum, message, args, **kwargs)
        def logToRoot(message, *args, **kwargs):
            logging.log(levelNum, message, *args, **kwargs)
    
        logging.addLevelName(levelNum, levelName)
        setattr(logging, levelName, levelNum)
        setattr(logging.getLoggerClass(), methodName, logForLevel)
        setattr(logging, methodName, logToRoot)
    
    
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2023 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from builtins import NotImplementedError

from holado.common.handlers.undefined import default
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tables.table_with_header import TableWithHeader
from holado_core.common.tables.table_row import TableRow
import abc
from holado_core.common.tools.tools import Tools
from holado_core.common.tables.table import Table
from holado_db.tools.db.query.base.query_builder import QueryBuilder
from holado_data.data.generator.base import BaseGenerator
from holado_python.standard_library.typing import Typing
from holado.common.handlers.object import Object

logger = logging.getLogger(__name__)


class DBClient(Object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, name, connect_kwargs, with_cursor=True, auto_commit=True):
        super().__init__(name)
        
        self.__connect_kwargs = connect_kwargs
        self.__with_cursor = with_cursor
        self.__auto_commit = auto_commit
        
        self.__connection = None
        self.__query_builder = None
        self.__audit_manager = None
        self.__column_names_by_table = {}
    
    @property
    def query_builder(self) -> QueryBuilder:
        return self.__query_builder
    
    @query_builder.setter
    def query_builder(self, builder):
        self.__query_builder = builder
    
    @property
    def is_connected(self):
        return self.__connection is not None
    
    @property
    def connection(self):
        return self.__connection
    
    @property
    def cursor(self):
        if not self.__with_cursor:
            raise TechnicalException(f"DB client '{self.__name}' doesn't manage cursor")
        return self.__cursor
    
    @property
    def audit_manager(self):
        if self.__audit_manager is None:
            self.__audit_manager = self._new_audit_manager()
        return self.__audit_manager
    
    def _new_audit_manager(self):
        """Create a new audit manager.
        By default, use a audit manager with a single table of audit for whole DB and with triggers on audited tables.
        Override this method in order to define which audit manager to use.
        """
        raise NotImplementedError()
    
    def connect(self):
        try:
            self.__connection = self._connect(**self.__connect_kwargs)
        except Exception as exc:
            Tools.raise_same_exception_type(exc, f"[{self.name}] Failed to connect with parameters {self.__connect_kwargs}")
        if self.__with_cursor:
            self.__cursor = self.__connection.cursor()
    
    def _connect(self, **kwargs):
        raise NotImplementedError()
    
    def _verify_is_connected(self):
        if not self.is_connected:
            raise FunctionalException(f"DB Client '{self.name}' is not connected")
    
    # Manage queries
    
    def insert(self, table_name, data: dict, do_commit=None):
        """
        Insert given data.
        Parameter 'data' has to be a dictionary with keys equal to table column names.
        """
        query, values = self.query_builder.insert(table_name, data)
        return self.execute_query(query, *values, do_commit=do_commit)
    
    def update(self, table_name, data: dict, where_data: dict=None, where_compare_data: list=None, do_commit=None):
        """
        Update given data.
        Parameters 'data' and 'where_data' have to be dictionaries with keys equal to table column names.
        """
        query, values = self.query_builder.update(table_name, data, where_data=where_data, where_compare_data=where_compare_data)
        return self.execute_query(query, *values, do_commit=do_commit)
    
    def select(self, table_name, where_data: dict=None, where_compare_data: list=None, sql_return="*", order_bys: list=None, **kwargs):
        """
        Select by filtering on given where data.
        @param where_data: dictionary of (field_name, value) for simple where clauses.
        @param where_compare_data: list of tuples (field_name, operator, value) for where clauses comparing fields with values.
        """
        query, values = self.query_builder.select(table_name, where_data=where_data, where_compare_data=where_compare_data, sql_return=sql_return, order_bys=order_bys)
        return self.execute_query(query, *values, do_commit=False, **kwargs)
    
    def delete(self, table_name, where_data: dict=None, where_compare_data: list=None, do_commit=None):
        """
        Delete by filtering on given where data.
        Parameter 'where_data' has to be a dictionary with keys equal to table column names.
        """
        query, values = self.query_builder.delete(table_name, where_data=where_data, where_compare_data=where_compare_data)
        return self.execute_query(query, *values, do_commit=do_commit)
    
    def count(self, table_name, where_data: dict=None, where_compare_data: list=None):
        result = self.select(table_name, where_data=where_data, where_compare_data=where_compare_data, sql_return="count(*)")
        return result[0][0].content
        
    def execute_query(self, query, *args, **kwargs):
        sql = self.query_builder.to_sql(query)
        
        # Force do_commit to False for select queries
        do_commit = kwargs.pop('do_commit', None)
        
        return self.execute(sql, *args, do_commit=do_commit, **kwargs)
        
    def execute(self, sql, *args, **kwargs):
        # Manage specific parameters
        if self.is_query_type(sql, 'select'):
            do_commit = False
        else:
            do_commit = kwargs.pop('do_commit', None)
            if do_commit is None:
                do_commit = self.__auto_commit
        result_as_dict_list = kwargs.pop('result_as_dict_list', False)
        as_generator = kwargs.pop('as_generator', False)
        if as_generator and do_commit:
            raise TechnicalException(f"'do_commit=True' and 'as_generator=True' are incompatible")
        if as_generator and not result_as_dict_list:
            raise TechnicalException(f"'as_generator=True' is possible only with 'result_as_dict_list=True'")
            
        self._verify_is_connected()
        
        # Execute
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"[{self.name}] Executing SQL [{Tools.truncate_text(sql, 1000)}] (args: {Tools.truncate_text(args, 1000)} ; kwargs: {Tools.truncate_text(kwargs, 1000)})...")
        try:
            if args:
                self.cursor.execute(sql, args)
            elif kwargs:
                self.cursor.execute(sql, kwargs)
            else:
                self.cursor.execute(sql)
        except self._get_base_exception_type() as exc:
            self.rollback()
            raise TechnicalException(f"[{self.name}] Error while executing SQL [{sql}] (args: {args} ; kwargs: {kwargs})") from exc
        
        # Get result
        if self.cursor.description:
            field_names = [field[0] for field in self.cursor.description]
            
            if result_as_dict_list:
                class Cursor2DictGenerator(BaseGenerator):
                    def __init__(self, field_names, cursor):
                        super().__init__(name="DB cursor to dict generator")
                        self.__field_names = field_names
                        self.__cursor = cursor
                        self.__cursor_iter = iter(self.__cursor)
                    
                    def __next__(self):
                        row_values = next(self.__cursor_iter)
                        return dict(zip(self.__field_names, row_values))
                
                gen = Cursor2DictGenerator(field_names, self.cursor)
                if as_generator:
                    res = gen
                else:
                    res = [e for e in gen]
            else:
                res = TableWithHeader()
                res.header = TableRow(cells_content=field_names)
                
                row_values = self.cursor.fetchone()
                while row_values:
                    res.add_row(cells_content=row_values)
                    row_values = self.cursor.fetchone()
        elif self.cursor.rowcount > 0:
            res = self.cursor.rowcount
        else:
            res = None
        
        self.__log_sql_result(res, message=f"Executed SQL [{Tools.truncate_text(sql, 1000)}] (args: {Tools.truncate_text(args, 1000)} ; kwargs: {Tools.truncate_text(kwargs, 1000)})")
        
        # Manage commit
        if do_commit:
            self.do_commit_else_rollback()

        return res
    
    def is_query_type(self, sql_or_query, query_type):
        from sql_metadata.keywords_lists import QueryType
        import sql_metadata
        import pypika.queries
        
        if isinstance(sql_or_query, pypika.queries.QueryBuilder):
            sql = self.query_builder.to_sql(sql_or_query)
        else:
            sql = sql_or_query
        if not isinstance(sql, str):
            raise TechnicalException(f"Unexpected type '{Typing.get_object_class_fullname(sql)}' for parameter sql_or_query")
        sql = sql.lower()
        
        # Parsing query with sql_metadata can be costly, thus begin by fast and simple methods
        query_type_str = query_type.lower() if isinstance(query_type, str) else query_type.name.lower()
        for qt in ['select', 'insert', 'update', 'delete']:
            if sql.startswith(qt):
                return qt == query_type_str
        
        # Parse more complex queries
        if isinstance(query_type, str):
            query_type = QueryType[query_type.upper()]
        if not isinstance(query_type, QueryType):
            raise TechnicalException(f"Unmanage query_type of type '{Typing.get_object_class_fullname(query_type)}'")
        
        try:
            p = sql_metadata.Parser(sql)
            res = p.query_type == query_type
        except ValueError as exc:
            if "Not supported query type" in str(exc):
                res = False
            else:
                raise TechnicalException(f"Failed to define query type") from exc
        except Exception as exc:
            raise TechnicalException(f"Failed to define query type") from exc
        
        return res
    
    def set_or_update_json_key_value(self, table_name, field_name, json_key, json_value, where_data: dict=None, where_compare_data: list=None):
        """
        Set or update a json field with key=value.
        """
        raise NotImplementedError()
        
    def _get_sql_placeholder(self):
        """
        Return the character/string to use as placeholder in SQL requests.
        """
        raise NotImplementedError()

    def _get_base_exception_type(self):
        raise NotImplementedError()
    
    def __log_sql_result(self, sql_result, message="SQL result", limit_rows=10):
        if Tools.do_log(logger, logging.DEBUG):
            res_str = self.__represent_sql_result(sql_result, limit_rows=limit_rows)
            if '\n' in res_str:
                logger.debug(f"[{self.name}] {message}:\n{Tools.indent_string(4, res_str)}")
            else:
                logger.debug(f"[{self.name}] {message} => {res_str}")
    
    def __represent_sql_result(self, sql_result, limit_rows = 10):
        if isinstance(sql_result, Table):
            return sql_result.represent(limit_rows=limit_rows)
        elif isinstance(sql_result, list) and limit_rows > 0 and len(sql_result) > limit_rows:
            return str(sql_result[:limit_rows])[:-1] + f", ...({len(sql_result)-limit_rows})]"
        else:
            return str(sql_result)
    
    
    # Manage transactions
    
    def commit(self):
        self.connection.commit()
        
    def rollback(self):
        self.connection.rollback()

    def do_commit_else_rollback(self, do_commit=True):
        if do_commit:
            try:
                self.commit()
            except self._get_base_exception_type() as exc:
                self.rollback()
                raise TechnicalException(f"[{self.name}] Error while commit after SQL [{sql}] (args: {args} ; kwargs: {kwargs})") from exc
        else:
            self.rollback()

    # Manage tables
    
    def exist_table(self, table_name):
        raise NotImplementedError()
    
    def create_table(self, table_name, create_sql, raise_if_exist=False, do_commit=True, do_audit=False):
        if not self.exist_table(table_name):
            self.execute(create_sql, do_commit=do_commit)
            if not self.exist_table(table_name):
                raise TechnicalException(f"Failed to create table '{table_name}' with SQL request [{create_sql}]")
            
            # Audit table
            if do_audit:
                self.audit_manager.audit_table(table_name)
        elif raise_if_exist:
            raise FunctionalException(f"Table '{table_name}' already exists")
    
    def drop_table(self, table_name, raise_if_not_exist=False, do_commit=True, drop_indexes=True):
        # Drop table indexes before
        if drop_indexes:
            self.drop_table_indexes(table_name, do_commit=do_commit)

        if self.exist_table(table_name):
            sql = f"drop table {table_name};"
            self.execute(sql, do_commit=do_commit)
            if self.exist_table(table_name):
                raise TechnicalException(f"Failed to drop table '{table_name}' with SQL request [{sql}]")
        elif raise_if_not_exist:
            raise FunctionalException(f"Table '{table_name}' doesn't exist")
    
    def get_table_column_names(self, table_name):
        if table_name not in self.__column_names_by_table:
            self.__column_names_by_table[table_name] = self._get_table_column_names(table_name)
        return self.__column_names_by_table[table_name]
    
    def _get_table_column_names(self, table_name):
        sql = f"select * from {table_name} limit 1"
        try:
            self.cursor.execute(sql)
        except self._get_base_exception_type() as exc:
            self.rollback()
            raise TechnicalException(f"[{self.name}] Error while executing SQL [{sql}]") from exc
        
        if self.cursor.description:
            res = [field[0] for field in self.cursor.description]
        else:
            raise TechnicalException(f"Failed to get column names of table {table_name}")
        
        return res

    def exist_table_index(self, table_name, index_name):
        raise NotImplementedError()

    def create_table_indexes(self, table_name, indexes_column_names, unique=False, if_not_exists=False, raise_if_exist=False, do_commit=True):
        for index_column_names in indexes_column_names:
            self.create_table_index(table_name, index_column_names, unique=unique, if_not_exists=if_not_exists, raise_if_exist=raise_if_exist, do_commit=do_commit)

    def _get_default_index_name(self, table_name, column_names):
        return f"idx_{table_name}_{'_'.join(column_names)}"

    def create_table_index(self, table_name, column_names, index_name=default, unique=False, if_not_exists=False, raise_if_exist=False, do_commit=True):
        if index_name is default:
            index_name = self._get_default_index_name(table_name, column_names)
        if if_not_exists or not self.exist_table_index(table_name, index_name):
            query = self.query_builder.create_index(index_name, table_name, column_names, unique=unique, if_not_exists=if_not_exists)
            return self.execute_query(query, do_commit=do_commit)
        elif raise_if_exist:
            raise FunctionalException(f"Table '{table_name}' index '{index_name}' already exists")
        return None

    def drop_table_indexes(self, table_name, index_names=None, if_exists=False, raise_if_not_exist=False, do_commit=True):
        if index_names is None:
            index_names = self.get_table_index_names(table_name)

        for index_name in index_names:
            self.drop_table_index(table_name, index_name, if_exists=if_exists, raise_if_not_exist=raise_if_not_exist, do_commit=do_commit)

    def drop_table_index(self, table_name, index_name, if_exists=False, raise_if_not_exist=False, do_commit=True):
        if if_exists or self.exist_table_index(table_name, index_name):
            query = self.query_builder.drop_index(index_name, if_exists=if_exists)
            return self.execute_query(query, do_commit=do_commit)
        elif raise_if_not_exist:
            raise FunctionalException(f"Table '{table_name}' index '{index_name}' doesn't exist")
        return None

    def get_table_index_names(self, table_name):
        raise NotImplementedError()





#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado.holado_config import Config
import traceback
import sys
from holado_core.common.handlers.enums import AccessType
import logging
import time
from holado_python.standard_library.typing import Typing
from holado.common.handlers.undefined import default_value

logger = logging.getLogger(__name__)


def reversed_enumerate(l):
    return zip(range(len(l)-1, -1, -1), reversed(l))

def reversed_range(min_, max_=None, step=-1):
    if max_ is None:
        min_, max_ = 0, min_
    return range(max_-1, min_-1, step)


class Tools(object):
    
    @classmethod
    def get_indent_string(cls, indent):
        return " " * indent
    
    @classmethod
    def indent_string(cls, indent, txt, *, do_indent_first_line=True):
        ind_str = Tools.get_indent_string(indent)
        lines = txt.split("\n") if txt else []
        if do_indent_first_line:
            return ind_str + ("\n" + ind_str).join(lines)
        else:
            return ("\n" + ind_str).join(lines)

    @classmethod
    def truncate_text(cls, text, length=Config.message_truncate_length, truncated_suffix=default_value, is_length_with_suffix=False, is_suffix_with_truncated_length=True):
        """Truncate text if needed.
        Default truncate suffix is '[...(xxx)]' if is_suffix_with_truncated_length=True else '[...]'
        Note: length can be None, meaning no truncate to do
        """
        if length is not None and len(text) > length:
            if truncated_suffix is default_value:
                truncated_suffix = f"[...({len(text)-length})]" if is_suffix_with_truncated_length else "[...]"
            if truncated_suffix:
                if is_length_with_suffix:
                    return text[0 : length - len(truncated_suffix)] + truncated_suffix
                else:
                    return text[0 : length] + truncated_suffix
            else:
                return text[0 : length]
        else:
            return text
    
    @classmethod
    def represent_exception(cls, exc, tb=None, indent=0):
        if sys.version_info >= (3,10):
            if tb is not None:
                list_format = traceback.format_exception(exc, value=exc, tb=tb)
            else:
                list_format = traceback.format_exception(exc)
        else:
            list_format = traceback.format_exception(type(exc), exc, exc.__traceback__)
        return Tools.indent_string(indent, "".join(list_format))
    
    @classmethod
    def represent_object(cls, obj, indent=0, *, do_indent_first_line=True, full_details=False, access_type=AccessType.Public, max_level=3):
        return cls.__represent_object(obj, indent, do_indent_first_line, full_details, access_type, max_level, [])
    
    @classmethod
    def __represent_object(cls, obj, indent, do_indent_first_line, full_details, access_type, max_level, _internal):
        from holado_core.common.tools.converters.converter import Converter
        
        if Converter.is_primitive(obj) and not Converter.is_dict(obj) and not Converter.is_list(obj):
            # logger.print(f"+++++ Tools.__represent_object: is primitive")
            res = f"{obj} [{str(type(obj))}]" if full_details else f"{obj}"
        else:
            id_obj = id(obj)
            if id_obj in _internal:
                res = f"[RECURSION] (type: {str(type(obj))} ; id: {id_obj})" if full_details else f"[RECURSION]"
            else:
                _internal.append(id_obj)
                
                if Converter.is_dict(obj):
                    # logger.print(f"+++++ Tools.__represent_object: is dict")
                    res_list = [f"{str(type(obj))}({id_obj})"] if full_details else []
                    # keys = sorted(value.keys())
                    keys = list(obj.keys())
                    for key in keys:
                        # logger.print(f"+++++ Tools.__represent_object:     key: {key}")
                        val_str = cls.__represent_object(obj[key], 0, True, full_details, access_type, max_level-1, _internal)
                        if '\n' in val_str:
                            res_list.append(f"    {key}:")
                            res_list.append(Tools.indent_string(4, val_str))
                        else:
                            res_list.append(f"    {key}: {val_str}")
                    res = "\n".join(res_list)
                elif Converter.is_list(obj):
                    # logger.print(f"+++++ Tools.__represent_object: is list")
                    res_list = [f"{str(type(obj))}({id_obj})"] if full_details else ["["]
                    # keys = sorted(value.keys())
                    for el in obj:
                    # for index, el in enumerate(obj):
                    #     logger.print(f"+++++ Tools.__represent_object:     index: {index}")
                        el_str = cls.__represent_object(el, 0, True, full_details, access_type, max_level-1, _internal)
                        if '\n' in el_str:
                            res_list.append("    {")
                            res_list.append(Tools.indent_string(4, el_str))
                            res_list.append("    }")
                        else:
                            res_list.append(f"    {el_str}")
                    if not full_details:
                        res_list.append("]")
                    res = "\n".join(res_list)
                else:
                    if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"Representing object of type {Typing.get_object_class_fullname(obj)}")
                    attributes = Typing.get_object_attributes(obj, access_type=access_type)
                    # logger.print(f"+++++ Tools.__represent_object: {str(type(obj))} is object with {attributes=}")
                    if attributes:
                        res_list = [f"{Typing.get_object_class_fullname(obj)}({id_obj})"] if full_details else [f"{Typing.get_object_class_fullname(obj)}"]
                        if max_level > 0:
                            for name, value in attributes:
                                # logger.print(f"+++++ Tools.__represent_object:     attribute '{name}'")
                                res_list.append(f"    {name}: {cls.__represent_object(value, 4, False, full_details, access_type, max_level-1, _internal)}")
                        res = "\n".join(res_list)
                    else:
                        res = f"{obj} [{Typing.get_object_class_fullname(obj)}({id_obj})]" if full_details else f"{obj}"
                        # raise TechnicalException(f"Unmanaged representation of object of type '{Typing.get_object_class_fullname(obj)}'")
        
        return Tools.indent_string(indent, res, do_indent_first_line=do_indent_first_line)
    
    @classmethod
    def has_sub_kwargs(cls, kwargs: dict, prefix: str):
        for k in kwargs.keys():
            if k.startswith(prefix):
                return True
        return False
    
    @classmethod    
    def pop_sub_kwargs(cls, kwargs: dict, prefix: str):
        res = {}
        for k in list(kwargs.keys()):
            if k.startswith(prefix):
                res[k[len(prefix):]] = kwargs.pop(k)
        return res
    
    @classmethod
    def do_log_level(cls, log_level, max_log_level):
        """Define the log level to use: if log_level is greater than max_log_level, return max_log_level.
        Note: says differently, the used log level is possibly reduced to max_log_level so that, if logger has not max_log_level enabled, no log is done
        Usage: use this method coupled with a parameter 'log_level' to define context where this method must log with level 'log_level' instead of lower levels if enabled to
        """
        return min(log_level, max_log_level)

    @classmethod
    def do_log(cls, logger, log_level, max_log_level=logging.CRITICAL):
        return logger.isEnabledFor(cls.do_log_level(log_level, max_log_level))
    
    @classmethod
    def do_log_if_objects_are_different(cls, logger, log_level, obj1, obj2, max_log_level=logging.CRITICAL):
        try:
            return cls.do_log(logger, log_level, max_log_level) and (obj1 != obj2)
        except Exception as exc:
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"Assume a log is needed since usually a comparison fails on mismatching types and thus objects are different. Obtained error: {exc}")
            return True
    
    @classmethod
    def raise_same_exception_type(cls, exception, new_message, add_from=True):
        exc = type(exception)(new_message)
        if add_from:
            raise exc from exception
        else:
            raise exc
    
    @classmethod
    def timer_s(cls):
        return time.perf_counter()
    
    @classmethod
    def timer_ns(cls):
        return time.perf_counter_ns()
    

# For HolAdo debug only
# if Tools.do_log(logger, logging.DEBUG):
#     # logger.debug("Imported Tools", stack_info=True)
#     stack = traceback.extract_stack(limit=None)
#     stack_str = Tools.indent_string(4, "".join(traceback.format_list(stack)))
#     print(f"Imported Tools\n{stack_str}")



from enum import IntEnum


class AISMessageType(IntEnum):
    """
    @summary: AIS message type name
    """
    T1_POSITION_REPORT = 1
    T2_POSITION_REPORT = 2
    T3_POSITION_REPORT = 3
    T4_BASE_STATION_REPORT = 4
    T5_SHIP_STATIC_DATA = 5
    T6_ADDRESSED_BINARY_MESSAGE = 6
    T7_BINARY_ACKNOWLEGDE = 7
    T8_BINARY_BROADCAST_MESSAGE = 8
    T9_STANDARD_SAR_AIRCRAFT_POSITION_REPORT = 9
    T10_COORDINATED_UTC_INQUIRY = 10
    T11_BASE_STATION_REPORT = 11
    T12_ADDRESSED_SAFETY_RELATED_MESSAGE = 12
    T13_BINARY_ACKNOWLEGDE = 13
    T14_SAFETY_RELATED_BROADCAST_MESSAGE = 14
    T15_INTERROGATION = 15
    T16_ASSIGNED_MODE_COMMAND = 16
    T17_GNSS_BROADCAST_BINARY_MESSAGE = 17
    T18_STANDARD_CLASS_B_POSITION_REPORT = 18
    T19_EXTENDED_CLASS_B_POSITION_REPORT = 19
    T20_DATA_LINK_MANAGEMENT = 20
    T21_AIDS_TO_NAVIGATION_REPORT = 21
    T22_CHANNEL_MANAGEMENT = 22
    T23_GROUP_ASSIGNMENT_COMMAND = 23
    T24_STATIC_DATA_REPORT = 24
    T25_SINGLE_SLOT_BINARY_MESSAGE = 25
    T26_MULTIPLE_SLOT_BINARY_MESSAGE = 26
    T27_LONG_RANG_AIS_BROADCAST_MESSAGE = 27
    
    
    


#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.exceptions.technical_exception import TechnicalException


class FindBuilder(object):
    """ Find builder
    """
    
    def __init__(self):
        self.__default_context = None
        self.__default_parameters = None
    
    @property
    def default_context(self):
        """
        Get or create default context.
        This method is usually used to build the default context.
        @return Default find context
        """
        if self.__default_context is None:
            from holado_core.common.finders.tools.find_context import FindContext
            self.__default_context = FindContext.default()
        return self.__default_context
    
    def context(self, find_context=None, find_type=None, container=None, containers=None):
        """
        @param find_context Find context to update
        @param find_type Find type
        @param container Container
        @param containers Containers
        @return Updated find context
        """
        if find_context is None:
            find_context = self.default_context
            
        if container is not None:
            from holado_core.common.finders.tools.find_context import ContainerFindContext
            res = ContainerFindContext(find_context=find_context).with_container(container)
        elif containers is not None:
            from holado_core.common.finders.tools.find_context import ListContainersFindContext
            res = ListContainersFindContext(find_context=find_context).with_containers(containers)
        else:
            res = find_context
            
        if find_type is not None:
            res = res.with_find_type(find_type)
        
        return res
    

    @property
    def default_parameters(self):
        """
        Get or create default parameters.
        This method is usually used to build the default parameters.
        @return Default find parameters
        """
        if self.__default_parameters is None:
            from holado_core.common.finders.tools.find_parameters import FindParameters
            self.__default_parameters = FindParameters.default()
        return self.__default_parameters
    
    def parameters(self, find_parameters=None, raise_exception=None):
        """
        @param raise_exception Whether raise exceptions or not
        @return Find parameters
        """
        if find_parameters is None:
            res = self.default_parameters
        else:
            res = find_parameters
            
        if raise_exception is not None:
            res = res.with_raise(raise_exception)
            
        return res
    
    def parameters_with_raise(self):
        """
        @return Default find parameters with raises
        """
        return self.parameters(True)
    
    def parameters_without_raise(self):
        """
        @return Default find parameters without any raise
        """
        return self.parameters(False)
    
    def update_find_inputs(self, container, candidates, find_context, find_parameters):
        if container is None and candidates is None:
            from holado_core.common.finders.tools.find_context import ContainerFindContext, ListContainersFindContext
            if isinstance(find_context, ContainerFindContext) and find_context.container is not None:
                container = find_context.container
            elif isinstance(find_context, ListContainersFindContext) and find_context.containers is not None:
                candidates = find_context.containers
            else:
                raise TechnicalException("No container is defined")
                
        if find_context is None:
            find_context = self.context()
        if find_parameters is None:
            find_parameters = self.parameters()
        
        return container, candidates, find_context, find_parameters

    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import threading
import logging
import os
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_multitask.multithreading.context.thread_context import ThreadContext
from holado_multitask.multiprocessing.context.process_context import ProcessContext
from holado_core.common.tools.tools import Tools
from holado_python.standard_library.typing import Typing
from holado.common.handlers.undefined import default_context
from holado.common.context.session_context import SessionContext
import psutil
import signal

from holado_test.behave.context.behave_session_context import BehaveSessionContext

logger = logging.getLogger(__name__)


class MultitaskManager(object):
    """
    Manage resources used by threads/processes run by HolAdo.
    It provides a ThreadContext/ProcessContext for each thread/process.
    """
    
    SESSION_SCOPE_NAME = "Session"
    SCENARIO_SCOPE_NAME = "Scenario"

    @classmethod
    def get_scope_name(cls, scope=default_context):
        if scope is default_context:
            if BehaveSessionContext.has_instance_scenario_context():
                return MultitaskManager.SCENARIO_SCOPE_NAME
            else:
                return MultitaskManager.SESSION_SCOPE_NAME
        elif scope is not None:
            return scope
        else:
            return MultitaskManager.SESSION_SCOPE_NAME
    
    
    @classmethod
    def get_process_id(cls, process=None):
        if process is None:
            return os.getpid()
        else:
            return process.pid
    
    @classmethod
    def get_parent_process_id(cls, pid=None):
        if pid is None:
            return os.getppid()
        else:
            process = psutil.Process(pid)
            return process.ppid()
    
    @classmethod
    def get_children_process_ids(cls, pid=None, recursive=False):
        if pid is None:
            pid = os.getpid()
        
        process = psutil.Process(pid)
        children = process.children(recursive=recursive)
        return [c.pid for c in children]
    
    @classmethod
    def kill_process(cls, pid, sig=signal.SIGTERM, do_kill_children=True, recursively=True):
        try:
            process = psutil.Process(pid)
        except psutil.NoSuchProcess:
            return
        
        # Kill children
        if do_kill_children:
            children = process.children(recursive=recursively)
            for proc in children:
                try:
                    proc.send_signal(sig)
                except signal.SIGTERM:
                    pass
        
        # Kill process
        try:
            process.send_signal(sig)
        except signal.SIGTERM:
            pass
    
    @classmethod
    def has_thread_native_id(cls):
        from holado_multitask.multithreading.threadsmanager import ThreadsManager
        return ThreadsManager.has_native_id()
    
    @classmethod
    def get_thread_uid(cls, thread=None):
        tid = cls.get_thread_id(thread=thread)
        if cls.has_thread_native_id():
            return tid
        else:
            return (cls.get_process_id(), tid)
    
    @classmethod
    def get_thread_id(cls, thread=None, native=None):
        if thread is None:
            thread = threading.current_thread()
        if native is None:
            native = cls.has_thread_native_id()
        elif native and not cls.has_thread_native_id():
            raise TechnicalException(f"System doesn't support thread native id")
        
        if native:
            return thread.native_id
        else:
            return thread.ident
        
    
    def __init__(self):
        self.__process_lock = threading.Lock()
        self.__process_context_by_uname = {}
        self.__parent_thread_uid_by_process_uname = {}
        self.__process_id_by_uname = {}
        self.__process_uname_by_id = {}
        self.__thread_uid_by_process_uname = {}
        
        self.__thread_lock = threading.Lock()
        self.__main_thread_uid = self.get_thread_uid()
        self.__thread_context_by_uname = {}
        self.__parent_thread_uid_by_thread_uname = {}
        self.__thread_uid_by_uname = {}
        self.__thread_uname_by_uid = {}
        
        self.__related_thread_uid_by_uid = {}
        
        self.__feature_context_by_thread_uid = {}
    
    @property
    def is_main_thread(self):
        thread_uid = self.get_thread_uid()
        return thread_uid == self.__main_thread_uid
    
    @property
    def main_thread_uid(self):
        return self.__main_thread_uid
    
    
    ### Manage processes
    
    def prepare_process(self, name, update_parent=True):
        """
        Create a process context and return a unique name based to given name
        """
        with self.__process_lock:
            res = name
            name_index = 0
            while res in self.__process_context_by_uname:
                name_index += 1
                res = f"{name}_{name_index}"
                
            self.__process_context_by_uname[res] = ProcessContext(name, res)
            
            if update_parent:
                uid = self.get_thread_uid()
                self.__parent_thread_uid_by_process_uname[res] = uid
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"New process context: name='{name}' ; unique name='{res}'{{}}".format(f" ; parent thread UID={uid}" if update_parent else ''))
        return res
    
    def set_process_id(self, name, pid):
        # logger.print(f"+++++++++ __parent_thread_uid_by_thread_uname: {self.__parent_thread_uid_by_thread_uname}")
        # logger.print(f"+++++++++ __feature_context_by_thread_uid: {self.__feature_context_by_thread_uid}")
        # logger.print(f"+++++++++ __thread_context_by_uname: {self.__thread_context_by_uname}")
        self.__process_id_by_uname[name] = pid
        self.__process_uname_by_id[pid] = name
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Set process PID: name='{name}' ; PID={pid}")
    
    def set_process_thread_uid(self, name, thread_uid):
        # Set main thread UID of process
        self.__thread_uid_by_process_uname[name] = thread_uid
        
        # Set parent thread of process as parent thread of process main thread 
        parent_thread_uid = self.__parent_thread_uid_by_process_uname[name]
        tuname = self._get_thread_unique_name_or_uid(thread_uid)
        self.__parent_thread_uid_by_thread_uname[tuname] = parent_thread_uid
        # logger.print(f"+++++++++ __parent_thread_uid_by_thread_uname: {self.__parent_thread_uid_by_thread_uname}")
        # logger.print(f"+++++++++ __feature_context_by_thread_uid: {self.__feature_context_by_thread_uid}")
        # logger.print(f"+++++++++ __thread_context_by_uname: {self.__thread_context_by_uname}")
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Set process thread UID: process name='{name}' ; thread name='{tuname}' ; thread UID={thread_uid} ; parent thread UID={parent_thread_uid}")
        
    def _get_process_unique_name_or_pid(self, pid=None):
        if pid is None:
            pid = self.get_process_id()
        if pid in self.__process_uname_by_id:
            return self.__process_uname_by_id[pid]
        else:
            # logger.debug(f"Using process PID {pid} as unique name")
            return pid
    
    
    ### Manage threads
    
    def prepare_thread(self, name, update_parent=True):
        """
        Create a thread context and return a unique name based to given name
        """
        with self.__thread_lock:
            res = name
            name_index = 0
            while res in self.__thread_context_by_uname:
                name_index += 1
                res = f"{name}_{name_index}"
                
            self.__thread_context_by_uname[res] = ThreadContext(name, res)
            
            this_uid = self.get_thread_uid()
            # In some cases (ex: on missing thread context), thread uid is used as name, and it can't be its own parent
            if update_parent and res != this_uid:
                # logger.print(f"+++++ Set parent thread UID: {res} -> {this_uid}")
                self.__parent_thread_uid_by_thread_uname[res] = this_uid
                
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"New thread context: name='{name}' ; unique name='{res}'{{}}".format(f" ; parent thread UID={this_uid}" if update_parent and res != this_uid else ''))
        return res
    
    def set_thread_uid(self, name, uid=None):
        if uid is None:
            uid = self.get_thread_uid()
        # logger.print(f"+++++ Set thread UID: {name} -> {uid}")
        self.__thread_uid_by_uname[name] = uid
        self.__thread_uname_by_uid[uid] = name
        self.__thread_context_by_uname[name].thread_uid = uid
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"Set thread UID: name='{name}' ; UID={uid}")
            
    def _get_thread_unique_name_or_uid(self, uid=None):
        if uid is None:
            uid = self.get_thread_uid()
        if uid in self.__thread_uname_by_uid:
            return self.__thread_uname_by_uid[uid]
        else:
            # logger.debug(f"Using thread UID {uid} as unique name")
            return uid
        
    def relate_thread_to(self, from_uid, to_uid):
        self.__related_thread_uid_by_uid[from_uid] = to_uid
    
    
    
    ### Manage process context
    
    def get_process_context(self, pid=None):
        if pid is None:
            pid = self.get_process_id()
        uname = self._get_process_unique_name_or_pid(pid=pid)
        # logger.print(f"++++ get_process_context: uid={uid} ; uname={uname}", stack_info=True)
        
        if uname not in self.__process_context_by_uname:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("Creating missing process context for unique name {}".format(f"'{uname}'" if isinstance(uname, str) else f"{uname}"))
            uname = self.prepare_process(uname, update_parent=False)
            self.set_process_id(uname, pid)
            
        return self.__process_context_by_uname[uname]
    
    
    
    
    ### Manage thread context
    
    def get_thread_context(self, uid=None):
        if uid is None:
            uid = self.get_thread_uid()
        uname = self._get_thread_unique_name_or_uid(uid=uid)
        # logger.print(f"++++ get_thread_context: uid={uid} ; uname={uname}", stack_info=True)
        
        if uname not in self.__thread_context_by_uname:
            if Tools.do_log(logger, logging.DEBUG):
                logger.debug("Creating missing thread context for unique name {}".format(f"'{uname}'" if isinstance(uname, str) else f"{uname}"))
            uname = self.prepare_thread(uname, update_parent=False)
            self.set_thread_uid(uname, uid)
            
        res = self.__thread_context_by_uname[uname]
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Thread context for tread UID '{uid}' (uname: '{uname}'): {res}")
        return res
   
    def get_parent_thread_context(self, uid=None):
        if uid is None:
            uid = self.get_thread_uid()
        uname = self._get_thread_unique_name_or_uid(uid=uid)
        if uname in self.__parent_thread_uid_by_thread_uname:
            parent_uid = self.__parent_thread_uid_by_thread_uname[uname]
            # logger.print(f"++++ get_parent_thread_context: {uid} -> {parent_uid}", stack_info=True)
            return self.get_thread_context(uid=parent_uid)
        else:
            return None
    
    def __find_thread_context_having(self, has_func, uid=None, do_log=False, _internal=None):
        if uid is None:
            uid = self.get_thread_uid()
        if _internal is None:
            _internal = []
        
        res = None
        while True:
            if uid in _internal:
                break
            else:
                _internal.append(uid)
                
            res = self.__find_thread_context_having__throw_thread_parents(has_func, uid=uid, do_log=do_log, _internal=_internal)
            if res is None:
                if uid in self.__related_thread_uid_by_uid:
                    # Try with related thread
                    old_uid = uid
                    uid = self.__related_thread_uid_by_uid[uid]
                    if do_log and Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[MultitaskManager.__find_thread_context_having] Not found in thread uid '{old_uid}', try in thread uid '{uid}'")
                else:
                    if do_log and Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[MultitaskManager.__find_thread_context_having] Not found in thread uid '{uid}', stop search")
                    return None
            else:
                if do_log and Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"[MultitaskManager.__find_thread_context_having] Found in thread uid '{uid}' => {res}")
                break
        
        return res
    
    def __find_thread_context_having__throw_thread_parents(self, has_func, uid=None, do_log=False, _internal=None):
        if uid is None:
            uid = self.get_thread_uid()
            
        # logger.print(f"+++++ Find thread context having feature context: uid={uid}")
        while True:
            uname = self._get_thread_unique_name_or_uid(uid=uid)
            if uname in self.__thread_context_by_uname:
                # logger.print(f"+++++ Find thread context having feature context: {uname} not in __thread_context_by_uname => None")
                # return None
                
                thread_context = self.__thread_context_by_uname[uname]
                found = False
                if isinstance(has_func, str):
                    if hasattr(thread_context, has_func):
                        found = getattr(thread_context, has_func)()
                else:
                    found = has_func(thread_context)
                if found:
                    # logger.print(f"+++++ Find thread context having feature context => found from uid={uid}")
                    if do_log and Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[MultitaskManager.__find_thread_context_having__throw_thread_parents] Found in thread uid '{uid}' (uname: '{uname}') => {thread_context}")
                    return thread_context
                else:
                    if do_log and Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                        logger.trace(f"[MultitaskManager.__find_thread_context_having__throw_thread_parents] Thread uid '{uid}' (uname: '{uname}') has a context, but it doesn't verify {has_func}")
            
            if uname in self.__parent_thread_uid_by_thread_uname:
                # Retry with parent thread uid
                uid = self.__parent_thread_uid_by_thread_uname[uname]
                if do_log and Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"[MultitaskManager.__find_thread_context_having__throw_thread_parents] Try in parent thread uid '{uid}'")
                # logger.print(f"+++++ Find thread context having feature context: in parent uid={uid}")
                return self.__find_thread_context_having(has_func, uid=uid, do_log=do_log, _internal=_internal)
            else:
                # logger.print(f"+++++ Find thread context having feature context: {uname} not in __parent_thread_uid_by_thread_uname => None")
                if do_log and Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                    logger.trace(f"[MultitaskManager.__find_thread_context_having__throw_thread_parents] Not found in thread uid '{uid}' (uname: '{uname}'), and it has not parent thread")
                return None
        
    
    
    ### Manage feature context
    
    def __update_feature_context_for_thread_uid(self, uid):
        if uid not in self.__feature_context_by_thread_uid:
            thread_context = self.__find_thread_context_having(ThreadContext.has_feature_context, uid)
            if thread_context is None:
                feature_context = None
            else:
                feature_context = thread_context.get_feature_context()
            # logger.print(f"+++++ Set feature context for thread UID {uid}: {feature_context}")
            is_reference = (thread_context.thread_uid == uid)
            self.__feature_context_by_thread_uid[uid] = (feature_context, is_reference)
            if Tools.do_log(logger, logging.DEBUG):
                # logger.debug(f"Updated feature context for thread: UID='{uid}' ; feature context={feature_context}")
                logger.debug(f"Updated feature context for thread: UID='{uid}' ; feature context={feature_context}   (registered list: {self.__feature_context_by_thread_uid})")
        
    def __remove_feature_context_for_any_thread_uid(self, feature_context):
        list_uid = list(self.__feature_context_by_thread_uid.keys())
        for uid in list_uid:
            if self.__feature_context_by_thread_uid[uid][0] is feature_context:
                del self.__feature_context_by_thread_uid[uid]
                if Tools.do_log(logger, logging.DEBUG):
                    logger.debug(f"Removed feature context ({id(feature_context)}) association with thread UID {uid}")
        
    def has_feature_context(self, is_reference=True, do_log=False):
        uid = self.get_thread_uid()
        # self.__update_feature_context_for_thread_uid(uid)
        # return self.__feature_context_by_thread_uid[uid] is not None
        
        if uid in self.__feature_context_by_thread_uid:
            res = self.__feature_context_by_thread_uid[uid][0] is not None
            if res and is_reference is not None:
                res &= (is_reference == self.__feature_context_by_thread_uid[uid][1])
            if do_log and Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"[MultitaskManager.has_feature_context] Thread uid '{uid}' is in __feature_context_by_thread_uid and is_reference == {is_reference} => {res}")
            return res
        
        thread_context = self.__find_thread_context_having(ThreadContext.has_feature_context, uid, do_log=do_log)
        res = thread_context is not None
        if res and is_reference is not None:
            res &= (thread_context.thread_uid == uid)
        
        return res
    
    def get_feature_context(self):
        uid = self.get_thread_uid()
        self.__update_feature_context_for_thread_uid(uid)
        res = self.__feature_context_by_thread_uid[uid][0]
        if logger.isEnabledFor(logging.TRACE):  # @UndefinedVariable
            logger.trace(f"Feature context for tread UID '{uid}': {res}")
        return res
    
    def set_feature_context(self, feature_context, is_reference=True):
        thread_context = self.get_thread_context()
        thread_context.set_feature_context(feature_context)
        
        self.__feature_context_by_thread_uid[thread_context.thread_uid] = (feature_context, is_reference)
        
        if Tools.do_log(logger, logging.DEBUG):
            tuname = thread_context.thread_unique_name
            tuname_txt = f"'{tuname}'" if isinstance(tuname, str) else f"{tuname}"
            tuid = thread_context.thread_uid
            logger.debug(f"Set feature context: thread unique name={tuname_txt} ; thread uid={tuid} ; feature context={feature_context} ; is reference={is_reference}")
    
    def delete_feature_context(self):
        tuname = self._get_thread_unique_name_or_uid()
        if tuname not in self.__thread_context_by_uname:
            raise TechnicalException(f"No thread context for thread unique name '{tuname}' (type: {Typing.get_object_class_fullname(tuname)})")
        
        # Remove any reference to feature context before delete it
        feature_context = self.__thread_context_by_uname[tuname].get_feature_context()
        if Tools.do_log(logger, logging.DEBUG):
            tuname_txt = f"'{tuname}'" if isinstance(tuname, str) else f"{tuname}"
            logger.debug(f"Deleting feature context: thread unique name={tuname_txt} ; feature context={feature_context}")
        self.__remove_feature_context_for_any_thread_uid(feature_context)
        
        # Delete feature context
        self.__thread_context_by_uname[tuname].delete_feature_context()
    
    
    
    


# -*- coding: utf-8 -*-

import logging
from holado_test.behave.behave import execute_steps

logger = logging.getLogger(__name__)


class TSConfigManager:

    def __init__(self):
        self.__func_path_manager = None
        
    def initialize(self, func_path_manager):
        self.__func_path_manager = func_path_manager
    
    @property
    def __path_manager(self):
        return self.__func_path_manager()
    
    def configure_system_with_default_settings(self):
        # Remove possible side effects of previous execution 
        execute_steps(u"""
            # Given ensure XXX
            """)
    








#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.tools.comparators.comparator import Comparator,\
    CompareOperator
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.tools.converters.converter import Converter
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)


class FloatComparator(Comparator):
    
    def __init__(self, diff_precision=None, relative_precision=1e-15):
        """
        Define the comparing method and its associated precision.
        If difference precision is defined, difference comparison is applied, else relative comparison is applied.
        By default, a relative comparison is used with a precision of 1e-15
        
        Notes: 
        1. Usually, python implement float type with 52 digits for significant digits, which is a precision of 2**-52=2.220446049250313e-16. 
           The default relative precision is 1e-15, a little bit over this limit precision, in order to manage small computational errors.
           In many cases, this precision is too small ; if you don't know which precision to use, a relative precision of 1e-9 is quite usual.
        2. For the same reason, the appropriate difference precision is dependent on the values compared and the number of comparable significant digits.
           If you are not sure, prefer the relative comparison. 
        :param diff_precision: If defined, a difference comparison is done with this precision
        :param relative_precision: If defined, a relative comparison is done with this precision
        """
        super().__init__("float")
        
        self.__diff_precision = diff_precision
        self.__relative_precision = relative_precision

        if self.__diff_precision is not None:
            self.__relative_precision = None        
        if self.__diff_precision is None and self.__relative_precision is None:
            raise TechnicalException("A precision must be defined") 
    
    def _convert_input(self, obj, name):
        if Converter.is_float(obj):
            res = float(obj)
        else:
            res = obj
            
        if not isinstance(res, float):
            raise FunctionalException(f"{name.capitalize()} value is not a float: [{res}] (type: {Typing.get_object_class_fullname(res)})")
        return res
    
    def _compare_result(self, obj_1, operator: CompareOperator, obj_2):
        if operator == CompareOperator.Different:
            return not self.__almost_equals(obj_1, obj_2)
        elif operator == CompareOperator.Equal:
            return self.__almost_equals(obj_1, obj_2)
        elif operator == CompareOperator.Inferior:
            return (obj_1 < obj_2)
        elif operator == CompareOperator.InferiorOrEqual:
            return (obj_1 < obj_2) or self.__almost_equals(obj_1, obj_2)
        elif operator == CompareOperator.Superior:
            return (obj_1 > obj_2)
        elif operator == CompareOperator.SuperiorOrEqual:
            return (obj_1 > obj_2) or self.__almost_equals(obj_1, obj_2)
        else:
            raise TechnicalException(f"Unmanaged compare operator {operator}")
    
    def __almost_equals(self, val_1, val_2):
        if self.__relative_precision is not None:
            return self.__almost_equals_relative(val_1, val_2)
        else:
            return self.__almost_equals_diff(val_1, val_2)
    
    def __almost_equals_diff(self, val_1, val_2):
        diff = abs(val_1 - val_2)
        return (diff < self.__diff_precision)
    
    def __almost_equals_relative(self, val_1, val_2):
        diff = abs(val_1 - val_2)
        vmax = max(abs(val_1), abs(val_2))
        if vmax > 0:
            return (diff < vmax * self.__relative_precision)
        else:
            return True
    
    
    




#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2023 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
import json
from holado_core.common.exceptions.functional_exception import FunctionalException
import os

logger = logging.getLogger(__name__)



class ExecutionHistoricManager(object):
    def __init__(self, report_path):
        self.__report_path = report_path
        self.__report_execution_historic = []
        
        self.__import_report_execution_historic()
    
    @property
    def report_execution_historic(self):
        return self.__report_execution_historic
    
    def __import_report_execution_historic(self):
        features_path = os.path.join(self.__report_path, "Features")
        if os.path.exists(features_path):
            lp = sorted(os.listdir(features_path))
            for cp in lp:
                cur_feature_path = os.path.join(features_path, cp)
                if os.path.isdir(cur_feature_path):
                    feature_execution_historic = self.__extract_feature_execution_historic(cur_feature_path)
                    if feature_execution_historic:
                        self.__report_execution_historic.append(feature_execution_historic)
        else:
            raise FunctionalException(f"No feature in report path '{self.__report_path}'")
    
    def __extract_feature_execution_historic(self, feature_path):
        res = None
        
        scenarios_path = os.path.join(feature_path, "Scenarios")
        if os.path.exists(scenarios_path):
            lp = sorted(os.listdir(scenarios_path))
            
            # Extract feature info
            if len(lp) > 1:
                last_scenario_path = os.path.join(scenarios_path, lp[-2])
                scenario_execution_historic = self.__extract_scenario_execution_historic(last_scenario_path)
                res = {'feature': scenario_execution_historic[0]['feature'],
                       'scenarios': []}
            
            for cp in lp:
                cur_scenario_path = os.path.join(scenarios_path, cp)
                if os.path.isdir(cur_scenario_path):
                    scenario_execution_historic = self.__extract_scenario_execution_historic(cur_scenario_path)
                    if scenario_execution_historic:
                        res['scenarios'].extend(scenario_execution_historic[0]['scenarios'])
        
        return res
    
    def __extract_scenario_execution_historic(self, scenario_path):
        filepath = os.path.join(scenario_path, "execution_historic.json")
        with open(filepath, "r") as fin:
            return json.load(fin)
        
    def extract_execution_historic_data(self, execution_historic, data_fullnames):
        res = []
        
        for fullname in data_fullnames:
            data = self.__extract_object_data(execution_historic, fullname)
            if fullname.endswith('tags') and len(data) > 0:
                data = "-t " + "-t ".join(data)
            res.append((fullname, data))
        
        return res
    
    def __extract_object_data(self, obj, fullname):
        names = fullname.split('.', maxsplit=1)
        if names[0] in obj:
            res = obj[names[0]]
            if len(names) > 1:
                return self.__extract_object_data(res, names[1])
            else:
                return res
        else:
            raise FunctionalException(f"Not found data '{fullname}' in object: {obj}")
        


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: custom_types/field_types.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1e\x63ustom_types/field_types.proto\x12\x0c\x63ustom_types\"|\n\x0eMessageWithMap\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x32\n\x03map\x18\x02 \x03(\x0b\x32%.custom_types.MessageWithMap.MapEntry\x1a*\n\x08MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"]\n\x13MessageWithOptional\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04text\x18\x02 \x01(\t\x12\x1a\n\roptional_text\x18\x03 \x01(\tH\x00\x88\x01\x01\x42\x10\n\x0e_optional_text\"1\n\x13MessageWithRepeated\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0e\n\x06values\x18\x02 \x03(\x05\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'custom_types.field_types_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MESSAGEWITHMAP_MAPENTRY._options = None
  _MESSAGEWITHMAP_MAPENTRY._serialized_options = b'8\001'
  _globals['_MESSAGEWITHMAP']._serialized_start=48
  _globals['_MESSAGEWITHMAP']._serialized_end=172
  _globals['_MESSAGEWITHMAP_MAPENTRY']._serialized_start=130
  _globals['_MESSAGEWITHMAP_MAPENTRY']._serialized_end=172
  _globals['_MESSAGEWITHOPTIONAL']._serialized_start=174
  _globals['_MESSAGEWITHOPTIONAL']._serialized_end=267
  _globals['_MESSAGEWITHREPEATED']._serialized_start=269
  _globals['_MESSAGEWITHREPEATED']._serialized_end=318
# @@protoc_insertion_point(module_scope)



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of self software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and self permission notice shall be included in all copies or substantial portions of the Software.
#
# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

from holado_core.common.criterias.tools.criteria_context import CriteriaContext
from holado_core.common.criterias.tools.criteria_parameters import CriteriaParameters
import logging
from holado_ui.ui.gui.criterias.enums import CheckEditableModes
from holado_ui.ui.gui.criterias.gui_criteria import GUICriteria
import re
from holado_core.common.tools.tools import Tools
from holado.holado_config import Config
from holado_core.common.exceptions.technical_exception import TechnicalException
from holado_core.common.exceptions import technical_exception
from holado_python.standard_library.typing import Typing

logger = logging.getLogger(__name__)



class SeleniumCriteria(GUICriteria):
    """ Criteria for Selenium elements.
    """
    
    def __init__(self, selenium_inspector, check_editable:CheckEditableModes=CheckEditableModes.NoCheck):
        super().__init__(check_editable)
        
        from holado_ui_selenium.ui.gui.selenium.inspectors.selenium_inspector import SeleniumInspector
        if selenium_inspector is None:
            raise technical_exception(f"Parameter 'selenium_inspector' must be defined")
        if not isinstance(selenium_inspector, SeleniumInspector):
            raise technical_exception(f"Parameter 'selenium_inspector' is of type {Typing.get_object_class_fullname(selenium_inspector)}' that is not a SeleniumInspector")
        self.__inspector = selenium_inspector
        
        self.__attribute_value_by_attribute_name = {}
        self.__pattern_by_attribute_name = {}
        self.__content_text = None
        self.__pattern_content_text = None
        self.__pattern_inner_html = None
        self.__inner_text = None
        self.__pattern_inner_text = None
        self.__tag_name = None
        self.__pattern_tag_name = None
    
    def set_attribute(self, attr_name, attr_value):
        """
        @param attr_name Attribute name
        @param attr_value Attribute value
        """
        if attr_value is not None and len(attr_value) > 0:
            self.__attribute_value_by_attribute_name[attr_name] = attr_value
        elif attr_name in self.__attribute_value_by_attribute_name:
            del self.__attribute_value_by_attribute_name[attr_name]
    
    def get_attribute(self, attr_name):
        """
        @param attr_name Attribute name
        @return Attribute value
        """
        return self.__attribute_value_by_attribute_name.get(attr_name)
    
    def set_attribute_pattern(self, attr_name, attr_value, case_insensitive=False):
        """
        @param attr_name Attribute name
        @param attr_value Attribute value pattern
        @param case_insensitive If True set case insensitive flag
        """
        if attr_value is not None and len(attr_value) > 0:
            if case_insensitive:
                pattern = re.compile(attr_value, re.IGNORECASE)
            else:
                pattern = re.compile(attr_value)
            self.__pattern_by_attribute_name[attr_name] = pattern
        elif attr_name in self.__pattern_by_attribute_name:
            del self.__pattern_by_attribute_name[attr_name]
    
    @property
    def class_name(self):
        """
        @return Class name
        """
        return self.get_attribute("class")
    
    @class_name.setter
    def class_name(self, class_name):
        """
        @param class_name Class name
        """
        self.set_attribute("class", class_name)
    
    @property
    def class_name_pattern(self):
        """
        @return Class name
        """
        pattern = self.get_attribute_pattern("class")
        if pattern is not None:
            return pattern.pattern
        else:
            return None
    
    @class_name_pattern.setter
    def class_name_pattern(self, class_name):
        """
        @param class_name Class name
        """
        self.set_class_name_pattern(class_name)
    
    def set_class_name_pattern(self, class_name, case_insensitive=False):
        """
        @param class_name Class name
        """
        self.set_attribute_pattern("class", class_name, case_insensitive)
    
    @property
    def content_text(self):
        """
        @return Content text
        """
        return self.__content_text
    
    @content_text.setter
    def content_text(self, content_text):
        """
        @param content_text Content text
        """
        if content_text is not None and len(content_text) > 0:
            self.__content_text = content_text.decode('string_escape')
        else:
            self.__content_text = None
    
    @property
    def content_text_pattern(self):
        """
        @return Content text pattern
        """
        pattern = self.__pattern_content_text
        if pattern is not None:
            return pattern.pattern
        else:
            return None
    
    @content_text_pattern.setter
    def content_text_pattern(self, content_text):
        """
        @param content_text Content text pattern
        """
        self.set_content_text_pattern(content_text)
    
    def set_content_text_pattern(self, content_text, unescape=True):
        """
        @param content_text Content text pattern
        @param unescape If True unescape the content text pattern
        """
        if content_text is not None and len(content_text) > 0:
            if unescape:
                self.__pattern_content_text = re.compile(content_text.decode('string_escape'))
            else:
                self.__pattern_content_text = re.compile(content_text)
        else:
            self.__pattern_content_text = None
    
    @property
    def id(self):
        """
        @return Id
        """
        return self.get_attribute("id")
    
    @id.setter
    def id(self, id_):
        """
        @param id Id
        """
        self.set_attribute("id", id_)
    
    @property
    def id_pattern(self):
        """
        @return Id pattern
        """
        pattern = self.get_attribute_pattern("id")
        if pattern is not None:
            return pattern.pattern
        else:
            return None
    
    @id_pattern.setter
    def id_pattern(self, id_):
        """
        @param id Id pattern
        """
        self.set_attribute_pattern("id", id_)
    
    @property
    def inner_html_pattern(self):
        """
        @return Inner HTML pattern
        """
        pattern = self.__pattern_inner_html
        if pattern is not None:
            return pattern.pattern
        else:
            return None
    
    @inner_html_pattern.setter
    def inner_html_pattern(self, inner_html):
        """
        @param inner_html Inner HTML pattern
        """
        self.set_inner_html_pattern(inner_html)
    
    def set_inner_html_pattern(self, inner_html, unescape=True):
        """
        @param inner_html Inner HTML pattern
        @param unescape If True unescape the inner HTML pattern
        """
        if inner_html is not None and len(inner_html) > 0:
            if unescape:
                self.__pattern_inner_html = re.compile(inner_html.decode('string_escape'))
            else:
                self.__pattern_inner_html = re.compile(inner_html)
        else:
            self.__pattern_inner_html = None
    
    @property
    def inner_text(self):
        """
        @return Inner text
        """
        return self.__inner_text
    
    @inner_text.setter
    def inner_text(self, inner_text):
        """
        @param inner_text Inner text
        """
        if inner_text is not None and len(inner_text) > 0:
            self.__inner_text = inner_text.decode('string_escape')
        else:
            self.__inner_text = None
    
    @property
    def inner_text_pattern(self):
        """
        @return Inner text pattern
        """
        pattern = self.__pattern_inner_text
        if pattern is not None:
            return pattern.pattern
        else:
            return None
    
    @inner_text_pattern.setter
    def inner_text_pattern(self, inner_text):
        """
        @param inner_text Inner text pattern
        """
        self.set_inner_text_pattern(inner_text)
    
    def set_inner_text_pattern(self, inner_text, unescape=True):
        """
        @param inner_text Inner text pattern
        @param unescape If True unescape the content text pattern
        """
        if inner_text is not None and len(inner_text) > 0:
            if unescape:
                self.__pattern_inner_text = re.compile(inner_text.decode('string_escape'))
            else:
                self.__pattern_inner_text = re.compile(inner_text)
        else:
            self.__pattern_inner_text = None
    
    @property
    def name(self):
        """
        @return Name
        """
        return self.get_attribute("name")
    
    @name.setter
    def name(self, name):
        """
        @param name Name
        """
        self.set_attribute("name", name)
    
    @property
    def name_pattern(self):
        """
        @return Name pattern
        """
        pattern = self.get_attribute_pattern("name")
        if pattern is not None:
            return pattern.pattern
        else:
            return None
    
    @name_pattern.setter
    def name_pattern(self, name):
        """
        @param name Name pattern
        """
        self.set_attribute_pattern("name", name)
    
    @property
    def tag_name(self):
        """
        @return Tag name
        """
        return self.__tag_name
    
    @tag_name.setter
    def tag_name(self, tag_name):
        """
        @param tag_name Tag name
        """
        if tag_name is not None and len(tag_name) > 0:
            self.__tag_name = tag_name.decode('string_escape')
        else:
            self.__tag_name = None
    
    @property
    def tag_name_pattern(self):
        """
        @return Tag name pattern
        """
        pattern = self.__pattern_tag_name
        if pattern is not None:
            return pattern.pattern
        else:
            return None
    
    @tag_name_pattern.setter
    def tag_name_pattern(self, tag_name):
        """
        @param tag_name Tag name pattern
        """
        self.set_tag_name_pattern(tag_name)
    
    def set_tag_name_pattern(self, tag_name, unescape=True):
        """
        @param tag_name Tag name pattern
        @param unescape If True unescape the content text pattern
        """
        if tag_name is not None and len(tag_name) > 0:
            if unescape:
                self.__pattern_tag_name = re.compile(tag_name.decode('string_escape'))
            else:
                self.__pattern_tag_name = re.compile(tag_name)
        else:
            self.__pattern_tag_name = None
    
    def validate(self, element, criteria_context: CriteriaContext, criteria_parameters: CriteriaParameters):
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"{self._get_ident_string_level(criteria_parameters)}[{self} ; [{self._internal_api.get_element_description(element)}]] validate : begin")
        
        # Prepare analyze time spent
        if criteria_parameters.analyze_time_spent:
            start = Tools.timer_s()
        
        # Validate self element
        res = self.validate_element(element, criteria_context, criteria_parameters)

        # Verify element children
        if res and (self.__content_text is not None or self.__pattern_content_text is not None):
            # Verify that no child also verify the same criteria
            res = self.__no_child_validates(element, criteria_context, criteria_parameters)
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.debug(f"{self._get_indent_string_level(criteria_parameters)}[{self} ; [{self._internal_api.get_element_description(element)}]] validate : end -> {res}")
        
        # Analyze time spent
        if criteria_parameters.analyze_time_spent:
            duration = Tools.timer_s() - start
            if duration > Config.threshold_warn_time_spent_s:
                logger.warning(f"{self._get_indent_string_level(criteria_parameters)}[{self} ; [{self._internal_api.get_element_description(element)}]] validate -> {res} (took {duration} s)")
        
        return res
    
    
    def __no_child_validates(self, element, criteria_context, criteria_parameters):
        res = True

        for child in self._internal_api.find_children(None, element, criteria_context, criteria_parameters):
            if self.validate_element(child, criteria_context, criteria_parameters):
                res = False
                break
        
        if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
            logger.trace(f"{self._get_indent_string_level(criteria_parameters)}    [{self} ; [{self._internal_api.get_element_description(element)}]] no child validates : end -> {res}")
        return res
    
    def validate_element(self, element, criteria_context, criteria_parameters):
        if not element.is_web_element():
            return False

        element_text = None
        web_element = self._get_web_element(element)
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.trace(f"{self._get_indent_string_level(criteria_parameters)}    [{self} ; [{self._internal_api.get_element_description(element)}]] validate element : begin")

        res = super().validate_element(element, criteria_context, criteria_parameters)

        # Values
        if res and self.__tag_name is not None:
            res = (self.__tag_name == web_element.tag_name)
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{self._get_indent_string_level(criteria_parameters)}        [{self}] validate tag_name -> {res}")
        
        if res and len(self.__attribute_value_by_attribute_name) > 0:
            for attr_name, attr_value in self.__attribute_value_by_attribute_name.items():
                value = web_element.get_attribute(attr_name)
                if value is not None:
                    if attr_name in ["class", "style"]:
                        res = (attr_value in value)
                    else:
                        res = (attr_value == value)
                else:
                    res = False
                if not res:
                    break
            
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{self._get_indent_string_level(criteria_parameters)}        [{self}] validate attribute value -> {res}")
        
        # Patterns
        if res and self.__pattern_inner_html is not None:
            value = self._internal_api.get_inner_html(element)
            res = (self.__pattern_inner_html.search(value) is not None)
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{self._get_indent_string_level(criteria_parameters)}        [{self}] validate pattern on inner html (obtained: '{value}' ; expected: '{self.__pattern_inner_html.pattern}') -> {res}")
        
        if res and self.__pattern_tag_name is not None:
            value = web_element.tag_name
            res = (self.__pattern_tag_name.search(value) is not None)
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{self._get_indent_string_level(criteria_parameters)}        [{self}] validate pattern on tag name (obtained: '{value}' ; expected: '{self.__pattern_tag_name.pattern}') -> {res}")
        
        if res and len(self.__pattern_by_attribute_name) > 0:
            for attr_name, attr_pattern in self.__pattern_by_attribute_name.items():
                value = web_element.get_attribute(attr_name)
                if value is not None:
                    res = (attr_pattern.search(value) is not None)
                else:
                    res = False
                
                if not res:
                    break
            
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{self._get_indent_string_level(criteria_parameters)}        [{self}] validate pattern on attributes -> {res}")
        
        # Values using get_element_text (put at end for performance)
        if res and self.__content_text is not None:
            if element_text is None:
                element_text = self._internal_api.get_element_text(element, criteria_context, criteria_parameters)
            res = (self.__content_text == element_text)
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{self._get_indent_string_level(criteria_parameters)}        [{self}] validate content text (obtained: '{element_text}' ; expected: '{self.__content_text}') -> {res}")
        
        if res and self.__inner_text is not None:
            if element_text is None:
                element_text = self._internal_api.get_element_text(element, criteria_context, criteria_parameters)
            res = (self.__inner_text == element_text)
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{self._get_indent_string_level(criteria_parameters)}        [{self}] validate inner text (obtained: '{element_text}' ; expected: '{self.__inner_text}') -> {res}")
        
        # Patterns using get_element_text (put at end for performance)
        if res and self.__pattern_content_text is not None:
            if element_text is None:
                element_text = self._internal_api.get_element_text(element, criteria_context, criteria_parameters)
            res = (self.__pattern_content_text.search(element_text) is not None)
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{self._get_indent_string_level(criteria_parameters)}        [{self}] validate pattern on content text (obtained: '{element_text}' ; expected: '{self.__pattern_content_text.pattern}') -> {res}")
        
        if res and self.__pattern_inner_text is not None:
            if element_text is None:
                element_text = self._internal_api.get_element_text(element, criteria_context, criteria_parameters)
            res = (self.__pattern_inner_text.search(element_text) is not None)
            if Tools.do_log(logger, logging.TRACE):  # @UndefinedVariable
                logger.trace(f"{self._get_indent_string_level(criteria_parameters)}        [{self}] validate pattern on inner text (obtained: '{element_text}' ; expected: '{self.__pattern_inner_text.pattern}') -> {res}")
        
        if Tools.do_log(logger, logging.DEBUG):
            logger.trace(f"{self._get_indent_string_level(criteria_parameters)}    [{self} ; [{self._internal_api.get_element_description(element)}]] validate element : end -> {res}")
        return res
    
    def _is_editable(self, element):
        if element.is_web_element:
            web_element = self._get_web_element(element)
            is_readonly = (web_element.get_attribute("readonly") is not None)
            return not is_readonly and web_element.is_enabled()
        else:
            return False
    
    def __str__(self):
        res_list = [super().__str__()]
        
        # Values
        if self.content_text is not None:
            res_list.append(f"&content_text='{self.content_text}'")
        if self.inner_text is not None:
            res_list.append(f"&inner_text='{self.inner_text}'")
        if self.tag_name is not None:
            res_list.append(f"&tag_name='{self.tag_name}'")
        if not self.__attribute_value_by_attribute_name.empty():
            res_list.append("&attributes={")
            attr_list = []
            for key, value in self.__attribute_value_by_attribute_name.items():
                attr_list.append(f"'{key}':'{value}'")
            res_list.append(", ".join(attr_list))
            res_list.append("}")
        
        # Patterns
        if self.__pattern_content_text is not None:
            res_list.append(f"&content_text_pattern='{self.content_text_pattern}'")
        if self.__pattern_inner_html is not None:
            res_list.append(f"&inner_html_pattern='{self.inner_html_pattern}'")
        if self.__pattern_inner_text is not None:
            res_list.append(f"&inner_text_pattern='{self.inner_text_pattern}'")
        if self.__pattern_tag_name is not None:
            res_list.append(f"&tag_name_pattern='{self.tag_name_pattern}'")
        if not self.__pattern_by_attribute_name.empty():
            res_list.append("&attributes_patterns={")
            attr_list = []
            for key, value in self.__pattern_by_attribute_name.items():
                attr_list.append(f"'{key}':'{value}'")
            res_list.append(", ".join(attr_list))
            res_list.append("}")
        
        return "".join(res_list)
    
    @property
    def _inspector(self):
        return self.__inspector
    
    
    def _internal_api(self):
        return self._inspector.internal_api
    
    def _get_web_element(self, element_holder):
        if element_holder.is_web_element:
            return element_holder.element
        else:
            raise TechnicalException(f"Holded element is of type '{Typing.get_object_class_fullname(element_holder.element)}' rather than WebElement")
    
    



#################################################
# HolAdo (Holistic Automation do)
#
# (C) Copyright 2021-2025 by Eric Klumpp
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the “Software”), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

# The Software is provided “as is”, without warranty of any kind, express or implied, including but not limited to the warranties of merchantability, fitness for a particular purpose and noninfringement. In no event shall the authors or copyright holders be liable for any claim, damages or other liability, whether in an action of contract, tort or otherwise, arising from, out of or in connection with the software or the use or other dealings in the Software.
#################################################

import logging
from holado_core.common.exceptions.functional_exception import FunctionalException
from holado.common.context.session_context import SessionContext
from holado_grpc.api.rpc.grpc_client import GRpcClient


logger = logging.getLogger(__name__)

# Activation of logs in gRPC library (activated logs are thrown in stdout, it doesn't follow logging settings)
# os.environ["GRPC_VERBOSITY"] = "DEBUG"
# os.environ["GRPC_TRACE"] = "http"
# os.environ["GRPC_TRACE"] = "all"


class GRpcManager(object):
    """
    Manage gRPC features, agnostic to internal gRPC library.
    """
    
    def __init__(self):
        self.__func_grpc_services = None
        self.__func_protobuf_converter = None
        self.__func_protobuf_messages = None
    
    def initialize(self, func_grpc_services, func_protobuf_converter, func_protobuf_messages):
        self.__func_grpc_services = func_grpc_services
        self.__func_protobuf_converter = func_protobuf_converter
        self.__func_protobuf_messages = func_protobuf_messages
        
    @property
    def __grpc_services(self):
        return self.__func_grpc_services()
        
    def new_client(self, name, **kwargs):
        if name is None:
            name = "Undefined"
        endpoint = kwargs.pop("endpoint")
        
        service_fullname = kwargs.pop("service") if "service" in kwargs else None
        services_fullnames = kwargs.pop("services") if "services" in kwargs else None
        if service_fullname is not None:
            service_descriptors = [self.__grpc_services.get_service_descriptor(service_fullname)]
        elif services_fullnames is not None:
            service_descriptors = [self.__grpc_services.get_service_descriptor(sfn) for sfn in services_fullnames]
        else:
            service_descriptors = None
    
        # Proxy settings
        #kwargs['channel_options'] = (('grpc.enable_http_proxy', 0),('grpc.enable_https_proxy', 0),)
        
        if service_descriptors is not None:
            res = GRpcClient(name, endpoint=endpoint, service_descriptors=service_descriptors, **kwargs)
        else:
            import grpc
            try:
                res = GRpcClient(name, endpoint=endpoint, **kwargs)
            except grpc.RpcError as exc:
                if hasattr(exc, "details"):
                    details = str(exc.details)
                    if "status = StatusCode.UNIMPLEMENTED" in details and 'details = "Method not found: grpc.reflection.v1alpha.ServerReflection/ServerReflectionInfo"' in details:
                        raise FunctionalException(f"Endpoint '{endpoint}' doesn't manage reflection, thus step parameter 'service' or 'services' has to be specified")
                raise exc
        
        res.initialize(self.__func_grpc_services, self.__func_protobuf_converter, self.__func_protobuf_messages)
        
        return res
        
            